# -*- coding: utf-8 -*-
import sys
from io import StringIO

from django.apps import apps
from django.core.management import call_command
from django.test import TestCase


class ShowPermissionsTests(TestCase):
    def _run_command(self, *args, **kwargs):
        """
        Utility to run the command and return captured output.
        """
        out = StringIO()
        sys_stdout = sys.stdout
        sys.stdout = out
        try:
            call_command("show_permissions", *args, **kwargs)
        finally:
            sys.stdout = sys_stdout
        return out.getvalue()

    def _check_header_in_output(self, app_labels, model_verbose, output):
        """
        Accepts a list of app label variants and checks if any of them exists with the model_verbose in the output.
        """
        for app_label in app_labels:
            header = f"Permissions for {app_label} | {model_verbose}"
            if header in output:
                return
        raise AssertionError(
            f"None of the expected headers found in output. Tried: {[f'Permissions for {label} | {model_verbose}' for label in app_labels]}"
        )

    def test_should_list_permissions_for_all_apps_excluding_defaults(self):
        output = self._run_command(verbosity=3)
        auth_verbose = apps.get_app_config("auth").verbose_name
        user_verbose = apps.get_model("auth", "user")._meta.verbose_name
        self.assertNotIn(
            f"Permissions for {auth_verbose} | {user_verbose}",
            output,
            "Should not list auth permissions without --all flag",
        )
        self.assertNotIn("auth.add_user", output)

    def test_should_include_all_apps_with_flag(self):
        output = self._run_command("--all", verbosity=3)

        auth_config = apps.get_app_config("auth")
        user_verbose = apps.get_model("auth", "user")._meta.verbose_name
        self._check_header_in_output(
            [auth_config.verbose_name, auth_config.label], user_verbose, output
        )
        self.assertIn("auth.add_user", output)

        admin_config = apps.get_app_config("admin")
        for model in admin_config.get_models():
            model_verbose = model._meta.verbose_name
            self._check_header_in_output(
                [admin_config.verbose_name, admin_config.label], model_verbose, output
            )

    def test_should_filter_by_app_label(self):
        output = self._run_command("--app-label", "auth", verbosity=3)

        auth_config = apps.get_app_config("auth")
        for model in auth_config.get_models():
            model_verbose = model._meta.verbose_name
            self._check_header_in_output(
                [auth_config.verbose_name, auth_config.label], model_verbose, output
            )

        self.assertIn("auth.change_user", output)

    def test_should_filter_by_app_and_model(self):
        output = self._run_command("auth.user", verbosity=3)

        auth_config = apps.get_app_config("auth")
        user_verbose = apps.get_model("auth", "user")._meta.verbose_name
        self._check_header_in_output(
            [auth_config.verbose_name, auth_config.label], user_verbose, output
        )
        self.assertIn("auth.change_user", output)

    def test_should_raise_error_for_invalid_model(self):
        with self.assertRaisesMessage(
            Exception, "Content type not found for 'fakeapp.nosuchmodel'"
        ):
            self._run_command("fakeapp.nosuchmodel", verbosity=3)

    def test_should_return_permissions_for_test_model(self):
        if apps.is_installed("tests"):
            output = self._run_command("tests.samplemodel", verbosity=3)
            self.assertIn("tests.samplemodel", output.lower())
            self.assertIn("can add", output.lower())

    def test_should_raise_error_for_invalid_app_label(self):
        with self.assertRaisesMessage(
            Exception, 'No content types found for app label "noapp".'
        ):
            self._run_command("--app-label", "noapp", verbosity=3)



# -*- coding: utf-8 -*-
import socket
import sys
import warnings
from typing import List

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.db import DEFAULT_DB_ALIAS

from django_extensions.management.utils import signalcommand
from django_extensions.utils.deprecation import RemovedInNextVersionWarning
from django_extensions.settings import SQLITE_ENGINES, POSTGRESQL_ENGINES, MYSQL_ENGINES


class Command(BaseCommand):
    help = """Generates the SQL to create your database for you, as specified in settings.py
The envisioned use case is something like this:

    ./manage.py sqlcreate [--database=<databasename>] | mysql -u <db_administrator> -p
    ./manage.py sqlcreate [--database=<databasname>] | psql -U <db_administrator> -W
    """  # noqa: E501

    requires_system_checks: List[str] = []
    can_import_settings = True

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "-R",
            "--router",
            action="store",
            dest="router",
            default=DEFAULT_DB_ALIAS,
            help="Use this router-database other then defined in settings.py",
        )
        parser.add_argument(
            "--database",
            default=DEFAULT_DB_ALIAS,
            help=(
                "Nominates a database to run command for. "
                'Defaults to the "%s" database.'
            )
            % DEFAULT_DB_ALIAS,
        )
        parser.add_argument(
            "-D",
            "--drop",
            action="store_true",
            dest="drop",
            default=False,
            help="If given, includes commands to drop any existing user and database.",
        )

    @signalcommand
    def handle(self, *args, **options):
        database = options["database"]
        if options["router"] != DEFAULT_DB_ALIAS:
            warnings.warn(
                "--router is deprecated. You should use --database.",
                RemovedInNextVersionWarning,
                stacklevel=2,
            )
            database = options["router"]

        dbinfo = settings.DATABASES.get(database)
        if dbinfo is None:
            raise CommandError("Unknown database %s" % database)

        engine = dbinfo.get("ENGINE")
        dbuser = dbinfo.get("USER")
        dbpass = dbinfo.get("PASSWORD")
        dbname = dbinfo.get("NAME")
        dbhost = dbinfo.get("HOST")
        dbclient = socket.gethostname()

        # django settings file tells you that localhost should be specified by leaving
        # the DATABASE_HOST blank
        if not dbhost:
            dbhost = "localhost"

        if engine in SQLITE_ENGINES:
            sys.stderr.write(
                "-- manage.py migrate will automatically create a sqlite3 database file.\n"  # noqa: E501
            )
        elif engine in MYSQL_ENGINES:
            sys.stderr.write(
                """-- WARNING!: https://docs.djangoproject.com/en/dev/ref/databases/#collation-settings
-- Please read this carefully! Collation will be set to utf8_bin to have case-sensitive data.
"""  # noqa: E501
            )
            print("CREATE DATABASE %s CHARACTER SET utf8 COLLATE utf8_bin;" % dbname)
            print(
                "GRANT ALL PRIVILEGES ON %s.* to '%s'@'%s' identified by '%s';"
                % (dbname, dbuser, dbclient, dbpass)
            )
        elif engine in POSTGRESQL_ENGINES:
            if options["drop"]:
                print("DROP DATABASE IF EXISTS %s;" % (dbname,))
                if dbuser:
                    print("DROP USER IF EXISTS %s;" % (dbuser,))

            if dbuser and dbpass:
                print(
                    "CREATE USER %s WITH ENCRYPTED PASSWORD '%s' CREATEDB;"
                    % (dbuser, dbpass)
                )
                print(
                    "CREATE DATABASE %s WITH ENCODING 'UTF-8' OWNER \"%s\";"
                    % (dbname, dbuser)
                )
                print("GRANT ALL PRIVILEGES ON DATABASE %s TO %s;" % (dbname, dbuser))
            else:
                print(
                    "-- Assuming that unix domain socket connection mode is being used because\n"  # noqa: E501
                    "-- USER or PASSWORD are blank in Django DATABASES configuration."
                )
                print("CREATE DATABASE %s WITH ENCODING 'UTF-8';" % (dbname,))
        else:
            # CREATE DATABASE is not SQL standard, but seems to be supported by most.
            sys.stderr.write(
                "-- Don't know how to handle '%s' falling back to SQL.\n" % engine
            )
            print("CREATE DATABASE %s;" % dbname)
            print("GRANT ALL PRIVILEGES ON DATABASE %s to %s;" % (dbname, dbuser))



# -*- coding: utf-8 -*-
import datetime

from django.utils.translation import gettext_lazy as _
from mongoengine.document import Document
from mongoengine.fields import DateTimeField, IntField, StringField
from mongoengine.queryset import QuerySetManager

from django_extensions.mongodb.fields import (
    AutoSlugField,
    CreationDateTimeField,
    ModificationDateTimeField,
)


class TimeStampedModel(Document):
    """
    TimeStampedModel

    An abstract base class model that provides self-managed "created" and
    "modified" fields.
    """

    created = CreationDateTimeField()
    modified = ModificationDateTimeField()

    class Meta:
        abstract = True


class TitleSlugDescriptionModel(Document):
    """
    TitleSlugDescriptionModel

    An abstract base class model that provides title and description fields
    and a self-managed "slug" field that populates from the title.
    """

    title = StringField(max_length=255)
    slug = AutoSlugField(populate_from="title")
    description = StringField(blank=True, null=True)

    class Meta:
        abstract = True


class ActivatorModelManager(QuerySetManager):
    """
    ActivatorModelManager

    Manager to return instances of ActivatorModel:
        SomeModel.objects.active() / .inactive()
    """

    def active(self):
        """
        Return active instances of ActivatorModel:

        SomeModel.objects.active()
        """
        return super().get_queryset().filter(status=1)

    def inactive(self):
        """
        Return inactive instances of ActivatorModel:

        SomeModel.objects.inactive()
        """
        return super().get_queryset().filter(status=0)


class ActivatorModel(Document):
    """
    ActivatorModel

    An abstract base class model that provides activate and deactivate fields.
    """

    STATUS_CHOICES = (
        (0, _("Inactive")),
        (1, _("Active")),
    )
    status = IntField(choices=STATUS_CHOICES, default=1)
    activate_date = DateTimeField(
        blank=True, null=True, help_text=_("keep empty for an immediate activation")
    )
    deactivate_date = DateTimeField(
        blank=True, null=True, help_text=_("keep empty for indefinite activation")
    )
    objects = ActivatorModelManager()

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        if not self.activate_date:
            self.activate_date = datetime.datetime.now()
        super().save(*args, **kwargs)



# -*- coding: utf-8 -*-
from django.template import Context, Template, TemplateSyntaxError
from django.test import TestCase


class HighlightTagExceptionTests(TestCase):
    """Tests for highlight tag exceptions."""

    def setUp(self):
        self.ctx = Context()

    def test_should_raise_TemplateSyntaxError(self):
        content = """{% load highlighting %}
{% highlight %}
{% endhighlight %}
"""
        with self.assertRaisesRegex(
            TemplateSyntaxError, "'highlight' statement requires an argument"
        ):
            Template(content).render(self.ctx)


class HighlightTagTests(TestCase):
    """Tests for highlight tag."""

    def setUp(self):
        self.ctx = Context()

    def test_should_highlight_python_syntax_with_name(self):
        content = """{% load highlighting %}
{% highlight 'python' 'Excerpt: blah.py' %}
def need_food(self):
    print("Love is colder than death")
{% endhighlight %}"""
        expected_result = """<div class="predesc"><span>Excerpt: blah.py</span></div><div class="highlight"><pre><span></span><span class="k">def</span> <span class="w"> </span><span class="nf">need_food</span><span class="p">(</span><span class="bp">self</span><span class="p">):</span>
    <span class="nb">print</span><span class="p">(</span><span class="s2">&quot;Love is colder than death&quot;</span><span class="p">)</span>
</pre></div>"""
        result = Template(content).render(self.ctx)

        self.assertHTMLEqual(result, expected_result)

    def test_should_highlight_bash_syntax_without_name(self):
        content = """{% load highlighting %}
{% highlight 'bash' %}
echo "Hello $1"
{% endhighlight %}"""
        expected_result = """<div class="highlight"><pre><span></span><span class="nb">echo</span><span class="w"> </span><span class="s2">&quot;Hello </span><span class="nv">$1</span><span class="s2">&quot;</span>
</pre></div>"""

        result = Template(content).render(self.ctx)

        self.assertHTMLEqual(result, expected_result)


class ParseTemplateTests(TestCase):
    """Tests for parse_teplate filter."""

    def test_should_mark_html_as_safe(self):
        ctx = Context({"value": "<h1>Hello World</h1>"})
        content = """{% load highlighting %}
{{ value|parse_template }}
"""
        expected_result = """<h1>Hello World</h1>"""

        result = Template(content).render(ctx)

        self.assertHTMLEqual(result, expected_result)



# -*- coding: utf-8 -*-
import os
import shutil

from django.core.management import call_command
from django.test import TestCase
from io import StringIO
from tests import testapp_with_no_models_file

from unittest.mock import patch


JOBS_DIR = os.path.join(testapp_with_no_models_file.__path__[0], "jobs")
TIME_PERIODS = ["hourly", "daily", "weekly", "monthly", "yearly"]


class CreateJobsTestsMixin:
    def tearDown(self):
        super().tearDown()
        try:
            shutil.rmtree(JOBS_DIR)
        except OSError:
            pass


class CreateJobsExceptionsTests(CreateJobsTestsMixin, TestCase):
    @patch("sys.stderr", new_callable=StringIO)
    @patch(
        "django_extensions.management.commands.create_jobs._make_writeable",
        side_effect=OSError,
    )
    def test_should_print_error_notice_on_OSError(self, m__make_writeable, m_stderr):
        call_command("create_jobs", "testapp_with_no_models_file")

        self.assertRegex(
            m_stderr.getvalue(),
            r"Notice: Couldn't set permission bits on \S+ You're probably using an uncommon filesystem setup. No problem.",
        )


class CreateJobsTests(CreateJobsTestsMixin, TestCase):
    def test_should_create_jobs_directory_structure_silently(self):
        call_command("create_jobs", "testapp_with_no_models_file")

        self.assertTrue(os.path.exists(JOBS_DIR))

    @patch("sys.stdout", new_callable=StringIO)
    def test_should_create_jobs_directory_structure_and_print_SUCCESS_message(
        self, m_stdout
    ):
        call_command("create_jobs", "testapp_with_no_models_file", verbosity=2)

        self.assertTrue(os.path.exists(JOBS_DIR))
        for time_period in TIME_PERIODS:
            self.assertIn(
                "testapp_with_no_models_file/jobs/{}/__init__.py".format(time_period),
                m_stdout.getvalue(),
            )

    @patch("sys.stdout", new_callable=StringIO)
    def test_should_not_override_already_created_jobs_directory_structure_and_print_that_already_exists(
        self, m_stdout
    ):
        call_command("create_jobs", "testapp_with_no_models_file")
        sample_file_path = os.path.join(JOBS_DIR, "sample.py")
        TEST_COMMENT = "# test"
        with open(sample_file_path, "a") as f:
            f.write(TEST_COMMENT)

        call_command("create_jobs", "testapp_with_no_models_file", verbosity=2)

        self.assertTrue(os.path.exists(JOBS_DIR))
        self.assertIn(TEST_COMMENT, open(sample_file_path).read())
        for time_period in TIME_PERIODS:
            self.assertIn(
                "testapp_with_no_models_file/jobs/{}/__init__.py already exists".format(
                    time_period
                ),
                m_stdout.getvalue(),
            )



# -*- coding: utf-8 -*-
from django_extensions.management.jobs import HourlyJob

from unittest import mock


HOURLY_JOB_MOCK = mock.MagicMock()


class Job(HourlyJob):
    help = "My sample hourly job."

    def execute(self):
        HOURLY_JOB_MOCK()



# -*- coding: utf-8 -*-
import json
import os
import re
import tempfile
from contextlib import contextmanager
from io import StringIO

from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase


def assert_looks_like_dotfile(output):
    assert output.startswith("digraph model_graph {\n")
    assert output.endswith("}\n")
    assert "// Dotfile by Django-Extensions graph_models\n" in output
    assert "// Labels\n" in output
    assert "// Relations\n" in output


def assert_looks_like_jsonfile(output):
    assert '"created_at": ' in output
    assert '"cli_options": ' in output
    assert '"app_name": "django.contrib.auth"' in output
    assert "created_at" in json.loads(output)


@contextmanager
def temp_output_file(extension=""):
    """Create writable tempfile in filesystem and ensure it gets deleted"""
    tmpfile = tempfile.NamedTemporaryFile(suffix=extension, delete=False)
    tmpfile.close()
    yield tmpfile.name
    os.unlink(tmpfile.name)


class GraphModelsOutputTests(TestCase):
    def test_graph_models_no_output_options(self):
        # Given no output-related options, default to output a Dotfile
        stdout = StringIO()
        call_command("graph_models", all_applications=True, stdout=stdout)
        assert_looks_like_dotfile(stdout.getvalue())

    def test_graph_models_dot_option_to_stdout(self):
        # --dot set but --output not set
        stdout = StringIO()
        call_command("graph_models", all_applications=True, dot=True, stdout=stdout)
        assert_looks_like_dotfile(stdout.getvalue())

    def test_graph_models_dot_option_to_file(self):
        # --dot set and --output set
        stdout = StringIO()
        with temp_output_file(".dot") as tmpfname:
            call_command(
                "graph_models",
                all_applications=True,
                dot=True,
                output=tmpfname,
                stdout=stdout,
            )
            with open(tmpfname, "r") as outfile:
                foutput = outfile.read()
        assert_looks_like_dotfile(foutput)
        assert stdout.getvalue() == ""

    def test_graph_models_dot_extensions_to_file(self):
        # --dot not set and --output set
        stdout = StringIO()
        with temp_output_file(".dot") as tmpfname:
            call_command(
                "graph_models", all_applications=True, output=tmpfname, stdout=stdout
            )
            with open(tmpfname, "r") as outfile:
                foutput = outfile.read()
        assert_looks_like_dotfile(foutput)
        assert stdout.getvalue() == ""

    def test_graph_models_dot_option_trumps_json_file_extension(self):
        # --dot set and --output set to filename ending with .json
        # assert that --dot option trumps .json file extension
        stdout = StringIO()
        with temp_output_file(".json") as tmpfname:
            call_command(
                "graph_models",
                all_applications=True,
                dot=True,
                output=tmpfname,
                stdout=stdout,
            )
            with open(tmpfname, "r") as outfile:
                foutput = outfile.read()
        assert_looks_like_dotfile(foutput)
        assert stdout.getvalue() == ""

    def test_graph_models_json_option_to_stdout(self):
        # --json set but --output not set
        out = StringIO()
        call_command("graph_models", all_applications=True, json=True, stdout=out)
        output = out.getvalue()
        assert_looks_like_jsonfile(output)

    def test_graph_models_json_option_to_file(self):
        # --dot set and --output set
        stdout = StringIO()
        with temp_output_file(".json") as tmpfname:
            call_command(
                "graph_models",
                all_applications=True,
                json=True,
                output=tmpfname,
                stdout=stdout,
            )
            with open(tmpfname, "r") as outfile:
                foutput = outfile.read()
        assert_looks_like_jsonfile(foutput)
        assert stdout.getvalue() == ""

    def test_graph_models_pydot_without_file(self):
        # use of --pydot requires specifying output file
        with self.assertRaises(CommandError):
            call_command("graph_models", all_applications=True, pydot=True)

    def test_graph_models_pygraphviz_without_file(self):
        # use of --pygraphviz requires specifying output file
        with self.assertRaises(CommandError):
            call_command("graph_models", all_applications=True, pygraphviz=True)

    def test_graph_models_relation_fields_only(self):
        # use of --relation-fields-only ignores all non-relation-fields
        stdout = StringIO()
        call_command("graph_models", all_applications=True, stdout=stdout, json=True)
        with_no_flag = json.loads(stdout.getvalue())

        stdout = StringIO()
        call_command(
            "graph_models",
            all_applications=True,
            relation_fields_only=True,
            stdout=stdout,
            json=True,
        )
        with_flag = json.loads(stdout.getvalue())

        # delete all entries for fields that do not display relations
        for graph_idx, graph in reversed(list(enumerate(with_no_flag["graphs"]))):
            for model_idx, model in reversed(list(enumerate(graph["models"]))):
                for field_idx, field in reversed(list(enumerate(model["fields"]))):
                    if (
                        field["type"]
                        not in ["ForeignKey (id)", "AutoField", "OneToOneField (id)"]
                        and not field["primary_key"]
                    ):
                        del with_no_flag["graphs"][graph_idx]["models"][model_idx][
                            "fields"
                        ][field_idx]

        # assert that manually deleting all non-relation-fields is same as using the flag
        self.assertEqual(with_no_flag, with_flag)


def test_disable_abstract_fields_not_active():
    out = StringIO()
    call_command(
        "graph_models",
        "django_extensions",
        include_models=["AbstractInheritanceTestModelChild"],
        disable_abstract_fields=False,
        stdout=out,
    )

    output = out.getvalue()
    assert "my_field_that_my_child_will_inherit" in output


def test_disable_abstract_fields_active():
    out = StringIO()
    call_command(
        "graph_models",
        "django_extensions",
        include_models=["AbstractInheritanceTestModelChild"],
        disable_abstract_fields=True,
        stdout=out,
    )

    output = out.getvalue()
    assert "my_field_that_my_child_will_inherit" not in output


def test_exclude_models_hides_relationships():
    """Expose bug #1229 where excluded models appear in relationships.

    They are replaced with an underscore, but the relationship is still there.
    """
    out = StringIO()
    call_command(
        "graph_models",
        "django_extensions",
        exclude_models=["Personality", "Note"],
        stdout=out,
    )

    output = out.getvalue()
    assert "tests_testapp_models_Person -> tests_testapp_models_Name" in output
    assert "tests_testapp_models_Person -> _" not in output


def test_hide_edge_labels():
    out = StringIO()
    call_command(
        "graph_models",
        "django_extensions",
        all_applications=True,
        hide_edge_labels=True,
        stdout=out,
    )
    output = out.getvalue()
    assert not re.search(r'\[label=\"[a-zA-Z]+"\]', output)



# -*- coding: utf-8 -*-
from io import StringIO

from django.conf import settings
from django.contrib.sites.models import Site
from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase
from django.test.utils import override_settings

from unittest.mock import patch


class SetDefaultSiteTests(TestCase):
    """Tests for set_default_site command."""

    @override_settings(SITE_ID=321)
    def test_should_raise_CommandError_when_Site_object_does_not_exist(self):
        with self.assertRaisesRegex(
            CommandError, "Default site with pk=321 does not exist"
        ):
            call_command("set_default_site")

    @patch("django_extensions.management.commands.set_default_site.socket")
    def test_should_raise_CommandError_if_system_fqdn_return_None(self, m_socket):
        m_socket.getfqdn.return_value = None
        with self.assertRaisesRegex(CommandError, "Cannot find systems FQDN"):
            call_command("set_default_site", "--system-fqdn")

    def test_should_raise_CommandError_if_both_domain_and_set_as_system_fqdn_are_present(
        self,
    ):
        with self.assertRaisesRegex(
            CommandError, "The set_as_system_fqdn cannot be used with domain option."
        ):
            call_command("set_default_site", "--domain=foo", "--system-fqdn")

    @override_settings(
        INSTALLED_APPS=[
            app for app in settings.INSTALLED_APPS if app != "django.contrib.sites"
        ]
    )
    def test_should_raise_CommandError_Sites_framework_not_installed(self):
        with self.assertRaisesRegex(
            CommandError, "The sites framework is not installed."
        ):
            call_command("set_default_site", "--domain=foo", "--system-fqdn")

    @patch("sys.stdout", new_callable=StringIO)
    def test_should_print_Nothing_to_update(self, m_stdout):
        call_command("set_default_site")

        self.assertIn(
            "Nothing to update (need --name, --domain and/or --system-fqdn)\n",
            m_stdout.getvalue(),
        )

    @patch("django_extensions.management.commands.set_default_site.socket")
    def test_should_use_domain_as_name_if_system_fqdn_return_domain_and_name_is_not_provided(
        self, m_socket
    ):
        m_socket.getfqdn.return_value = "test.com"

        call_command("set_default_site", "--system-fqdn")
        result = Site.objects.get(pk=settings.SITE_ID)

        self.assertEqual(result.name, "test.com")
        self.assertEqual(result.domain, "test.com")

    @patch("django_extensions.management.commands.set_default_site.socket")
    def test_should_set_custom_nameif_system_fqdn_return_domain_and_name_is_provided(
        self, m_socket
    ):
        m_socket.getfqdn.return_value = "test.com"

        call_command("set_default_site", "--system-fqdn", "--name=foo")
        result = Site.objects.get(pk=settings.SITE_ID)

        self.assertEqual(result.name, "foo")
        self.assertEqual(result.domain, "test.com")

    def test_should_set_name_and_domain_if_provided(self):
        call_command("set_default_site", "--name=foo", "--domain=bar")
        result = Site.objects.get(pk=settings.SITE_ID)

        self.assertEqual(result.name, "foo")
        self.assertEqual(result.domain, "bar")

    def test_should_set_name_only(self):
        call_command("set_default_site", "--name=foo")
        result = Site.objects.get(pk=settings.SITE_ID)

        self.assertEqual(result.name, "foo")
        self.assertEqual(result.domain, "example.com")

    def test_should_set_domain_only(self):
        call_command("set_default_site", "--domain=bar")
        result = Site.objects.get(pk=settings.SITE_ID)

        self.assertEqual(result.name, "example.com")
        self.assertEqual(result.domain, "bar")

    def test_should_not_raise_if_sites_installed_through_appconfig(self):
        with self.modify_settings(
            INSTALLED_APPS={
                "append": "django.contrib.sites.apps.SitesConfig",
                "remove": "django.contrib.sites",
            }
        ):
            call_command("set_default_site", "--name=foo", "--domain=foo.bar")



# -*- coding: utf-8 -*-
from unittest import mock

from django.db import models
from django.test import TestCase
from tests.testapp.models import (
    DummyRelationModel,
    InheritedFromPostWithUniqField,
    PostWithUniqField,
    ReverseModel,
    SecondDummyRelationModel,
    ThirdDummyRelationModel,
)

from django_extensions.db.fields import UniqueFieldMixin


class UniqFieldMixinTestCase(TestCase):
    def setUp(self):
        class MockField(UniqueFieldMixin):
            def __init__(self, **kwargs):
                for key, value in kwargs.items():
                    setattr(self, key, value)

        self.uniq_field = MockField(
            attname="uniq_field",
            max_length=255,
            boolean_attr=True,
            non_boolean_attr="non_boolean_attr",
        )

        f_dummy = DummyRelationModel.objects.create()
        s_dummy = SecondDummyRelationModel.objects.create()
        t_dummy = ThirdDummyRelationModel.objects.create()

        post = PostWithUniqField.objects.create(
            uniq_field="test_uniq",
            common_field="first",
            another_common_field="second",
            many_to_one_field=f_dummy,
            one_to_one_field=s_dummy,
        )
        post.many_to_many_field.add(t_dummy)
        post.save()

        ReverseModel.objects.create(post_field=post)

        self.post = post

    def tearDown(self):
        PostWithUniqField.objects.all().delete()
        DummyRelationModel.objects.all().delete()
        SecondDummyRelationModel.objects.all().delete()
        ThirdDummyRelationModel.objects.all().delete()
        ReverseModel.objects.all().delete()

    def test_check_is_bool_boolean_attr(self):
        self.assertIsNone(self.uniq_field.check_is_bool("boolean_attr"))

    def test_check_is_bool_non_boolean_attr(self):
        with self.assertRaisesMessage(
            ValueError,
            "'non_boolean_attr' argument must be True or False",
        ):
            self.uniq_field.check_is_bool("non_boolean_attr")

    def test__get_fields_returns_list_of_tulpes(self):
        uniq_mixin_fields = UniqueFieldMixin._get_fields(PostWithUniqField)
        self.assertIsInstance(uniq_mixin_fields, list)
        for field in uniq_mixin_fields:
            self.assertIsInstance(field, tuple)

    def test__get_fields_returns_correct_fields(self):
        option_fields = PostWithUniqField._meta.get_fields()
        uniq_mixin_fields = [
            i[0] for i in UniqueFieldMixin._get_fields(PostWithUniqField)
        ]

        self.assertEqual(len(option_fields), 9)
        self.assertEqual(len(uniq_mixin_fields), 7)

        fields_to_be_excluded_from_uniq_mixin_fields = [
            f
            for f in option_fields
            if f.is_relation
            and not f.one_to_one
            and not (f.many_to_one and f.related_model)
        ]

        for field in fields_to_be_excluded_from_uniq_mixin_fields:
            self.assertNotIn(field, uniq_mixin_fields)

    def test__get_fields_returns_correct_model(self):
        post_models = [i[1] for i in UniqueFieldMixin._get_fields(PostWithUniqField)]
        self.assertTrue(all(model is None for model in post_models))

        inherited_post_models = [
            i[1]
            for i in UniqueFieldMixin._get_fields(InheritedFromPostWithUniqField)
            if i[1]
        ]

        self.assertEqual(len(inherited_post_models), 6)
        self.assertTrue(
            all(model is PostWithUniqField) for model in inherited_post_models
        )

    def test_get_queryset(self):
        mocked_get_fields = ((models.CharField, PostWithUniqField),)
        with (
            mock.patch(
                "django_extensions.db.fields.UniqueFieldMixin._get_fields",
                return_value=mocked_get_fields,
            ),
            mock.patch(
                "tests.testapp.models.PostWithUniqField._default_manager.all"
            ) as mocked_qs_all,
        ):
            self.uniq_field.get_queryset(PostWithUniqField, models.CharField)

        mocked_qs_all.assert_called_with()

        mocked_get_fields = ((models.CharField, None),)

        with (
            mock.patch(
                "django_extensions.db.fields.UniqueFieldMixin._get_fields",
                return_value=mocked_get_fields,
            ),
            mock.patch(
                "tests.testapp.models.InheritedFromPostWithUniqField._default_manager.all"
            ) as mocked_qs_all,
        ):
            self.uniq_field.get_queryset(
                InheritedFromPostWithUniqField, models.CharField
            )

        mocked_qs_all.assert_called_with()

    def test_find_unique(self):
        def filter_func(*args, **kwargs):
            uniq_field = kwargs.get("uniq_field")
            if uniq_field == "a":
                return mocked_qs
            return None

        mocked_qs = mock.Mock(spec=PostWithUniqField.objects)
        mocked_qs.filter.side_effect = filter_func
        mocked_qs.exclude.return_value = mocked_qs

        field = models.CharField
        with mock.patch(
            "django_extensions.db.fields.UniqueFieldMixin.get_queryset",
            return_value=mocked_qs,
        ) as get_qs:
            res = self.uniq_field.find_unique(self.post, field, iter("abcde"))

        get_qs.assert_called_with(PostWithUniqField, field)
        mocked_qs.exclude.assert_called_with(pk=self.post.pk)
        self.assertEqual(res, "b")
        self.assertTrue(hasattr(self.post, "uniq_field"))
        self.assertEqual(self.post.uniq_field, "b")



# -*- coding: utf-8 -*-
from django.test import TestCase
from django.utils.text import Truncator

from django_extensions.admin import widgets

from .testapp import models


class ForeignKeySearchInputTestCase(TestCase):
    def test_widget_works(self):
        name = models.Name.objects.create(name="Name")
        person = models.Person.objects.create(
            name=name,
            age=30,
        )
        club = models.Club.objects.create(
            name="Club",
        )
        membership = models.Membership.objects.create(club=club, person=person)

        widget = widgets.ForeignKeySearchInput(
            models.Membership._meta.get_field("person").remote_field,
            ["person__name"],
        )

        label = widget.label_for_value(membership.pk)

        self.assertEqual(
            Truncator(person).words(14, truncate="..."),
            label,
        )

        # Just making sure rendering the widget doesn't cause any issue
        widget.render("person", person.pk)
        widget.render("person", None)

        # Check media to make sure rendering media doesn't cause any issue
        self.assertListEqual(
            [
                "/static/django_extensions/js/jquery.bgiframe.js",
                "/static/django_extensions/js/jquery.ajaxQueue.js",
                "/static/django_extensions/js/jquery.autocomplete.js",
            ],
            widget.media._js,
        )
        self.assertListEqual(
            ["/static/django_extensions/css/jquery.autocomplete.css"],
            list(widget.media._css["all"]),
        )



# -*- coding: utf-8 -*-
import json
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest
from django.core.management import CommandError, call_command
from django.db import connection
from django.db.migrations.recorder import MigrationRecorder

from django_extensions.management.commands.managestate import (
    DEFAULT_FILENAME,
    DEFAULT_STATE,
    Command,
)

pytestmark = [pytest.mark.django_db]

COMMAND = "managestate"
DEFAULT_FILE = Path(DEFAULT_FILENAME)


@pytest.fixture
def make_dump(request):
    request.addfinalizer(DEFAULT_FILE.unlink)
    return call_command(COMMAND, "dump", "-v", 0)


@pytest.fixture
def cmd_data():
    cmd = Command()
    cmd.verbosity = 0
    cmd.conn = connection
    data = cmd.get_migrated_apps()
    data.update(cmd.get_applied_migrations())
    return data


class TestManageStateExceptions:
    """Tests for managestate management command exceptions."""

    def test_bad_action(self):
        with pytest.raises(CommandError):
            call_command(COMMAND, "bad_action")

    def test_no_such_file(self):
        with pytest.raises(CommandError):
            call_command(COMMAND, "load", "-f", "non_existent_file")

    def test_not_a_file(self):
        with pytest.raises(CommandError), TemporaryDirectory() as tmp_dir:
            call_command(COMMAND, "load", "-f", tmp_dir)

    @pytest.mark.usefixtures("make_dump")
    def test_no_such_state(self):
        with pytest.raises(CommandError):
            call_command(COMMAND, "load", "non_existent_state")


class TestManageState:
    """Tests managestate management command"""

    def test_dump(self, request, capsys):
        request.addfinalizer(DEFAULT_FILE.unlink)
        call_command(COMMAND, "dump", "-v", 0)
        stdout, _ = capsys.readouterr()
        assert DEFAULT_FILE.exists()
        assert "successfully saved" in stdout

    @pytest.mark.parametrize("filename", [DEFAULT_FILENAME, "custom_f1l3n4m3.json"])
    def test_dump_files(self, request, filename):
        path = Path(filename)
        request.addfinalizer(path.unlink)
        call_command(COMMAND, "dump", "-f", filename)
        assert path.exists()

    @pytest.mark.parametrize("state", [DEFAULT_STATE, "custom_state"])
    def test_dump_states(self, request, state):
        request.addfinalizer(DEFAULT_FILE.unlink)
        call_command(COMMAND, "dump", state)
        with open(DEFAULT_FILE) as file:
            data = json.load(file)
        assert isinstance(data, dict)
        assert data.get(state) is not None

    @pytest.mark.usefixtures("make_dump")
    def test_load(self, capsys):
        call_command(COMMAND, "load", "-v", 0)
        stdout, _ = capsys.readouterr()
        assert "successfully applied" in stdout

    @pytest.mark.parametrize("filename", [DEFAULT_FILENAME, "custom_f1l3n4m3.json"])
    def test_load_files(self, request, capsys, filename):
        request.addfinalizer(Path(filename).unlink)
        call_command(COMMAND, "dump", "-f", filename)
        call_command(COMMAND, "load", "-f", filename)
        stdout, _ = capsys.readouterr()
        assert "successfully applied" in stdout

    @pytest.mark.parametrize("state", [DEFAULT_STATE, "custom_state"])
    def test_load_states(self, request, capsys, state):
        request.addfinalizer(DEFAULT_FILE.unlink)
        call_command(COMMAND, "dump", state)
        call_command(COMMAND, "load", state)
        stdout, _ = capsys.readouterr()
        assert "successfully applied" in stdout

    def test_migration_is_last_applied(self, cmd_data):
        migrations = MigrationRecorder(connection).applied_migrations()
        for app, migration in cmd_data.items():
            last_migration = sorted(filter(lambda x: x[0] == app, migrations))[-1][1]
            assert migration == last_migration



# -*- coding: utf-8 -*-
import importlib.util
import os
from io import StringIO

from django.core.management import CommandError, call_command
from django.test import TestCase
from django.test.utils import override_settings

from unittest import mock


class ResetDbExceptionsTests(TestCase):
    """Tests if reset_db command raises exceptions."""

    def test_should_raise_CommandError_when_database_does_not_exist(self):
        with self.assertRaisesRegex(
            CommandError, "Unknown database non-existing_database"
        ):
            call_command("reset_db", "--database=non-existing_database")

    @override_settings(
        DATABASES={
            "default": {
                "ENGINE": "django.db.backends.UNKNOWN",
                "NAME": "test.db",
            }
        }
    )
    def test_should_raise_CommandError_when_unknown_database_engine(self):
        with self.assertRaisesRegex(
            CommandError, "Unknown database engine django.db.backends.UNKNOWN"
        ):
            call_command("reset_db", "--noinput")

    @override_settings(
        DATABASES={
            "default": {
                "ENGINE": "django.db.backends.sqlite3",
            }
        }
    )
    def test_should_raise_CommandError_when_no_db_name_provided(self):
        with self.assertRaisesRegex(
            CommandError,
            "You need to specify DATABASE_NAME in your Django settings file.",
        ):
            call_command("reset_db", "--noinput")


@override_settings(
    DATABASES={
        "default": {
            "ENGINE": "django.db.backends.sqlite3",
            "NAME": "test_db.sqlite3",
        }
    }
)
class ResetDbSqlite3Tests(TestCase):
    """Tests for reset_db command and sqlite3 engine."""

    @mock.patch("sys.stdout", new_callable=StringIO)
    @mock.patch("django_extensions.management.commands.reset_db.input")
    def test_should_cancel_reset_db_if_input_is_different_than_yes(
        self, m_input, m_stdout
    ):
        m_input.return_value = "no"
        call_command("reset_db")
        self.assertEqual("Reset cancelled.\n", m_stdout.getvalue())

    @mock.patch("sys.stdout", new_callable=StringIO)
    @mock.patch.object(os, "unlink")
    def test_should_unlink_database_and_print_success_message(self, m_unlink, m_stdout):
        call_command("reset_db", "--noinput", verbosity=2)

        self.assertEqual("Reset successful.\n", m_stdout.getvalue())
        m_unlink.assert_called_once_with("test_db.sqlite3")

    @mock.patch("sys.stdout", new_callable=StringIO)
    @mock.patch.object(
        os,
        "unlink",
        side_effect=[
            OSError,
        ],
    )
    def test_should_print_successful_message_even_if_unlink_failed(
        self, m_unlink, m_stdout
    ):
        call_command("reset_db", "--noinput", verbosity=2)

        self.assertEqual("Reset successful.\n", m_stdout.getvalue())
        m_unlink.assert_called_once_with("test_db.sqlite3")


@override_settings(
    DATABASES={
        "default": {
            "ENGINE": "django.db.backends.mysql",
            "NAME": "test_db",
            "USER": "foo",
            "PASSWORD": "bar",
            "HOST": "127.0.0.1",
            "PORT": "",
        }
    }
)
class ResetDbMysqlTests(TestCase):
    """Tests for reset_db command and mysql engine."""

    @mock.patch("sys.stdout", new_callable=StringIO)
    @mock.patch("django_extensions.management.commands.reset_db.input")
    def test_should_cancel_reset_db_if_input_is_different_than_yes(
        self, m_input, m_stdout
    ):
        m_input.return_value = "no"

        call_command("reset_db")

        self.assertEqual("Reset cancelled.\n", m_stdout.getvalue())

    @mock.patch("sys.stdout", new_callable=StringIO)
    def test_should_drop_and_create_database_with_characterset_utf8_and_print_success_messsage(
        self, m_stdout
    ):
        m_database = mock.MagicMock()
        m_database.__spec__ = mock.Mock()
        m_connection = mock.Mock()
        m_database.connect.return_value = m_connection
        expected_calls = [
            mock.call("DROP DATABASE IF EXISTS `test_db`"),
            mock.call("CREATE DATABASE `test_db` CHARACTER SET utf8"),
        ]

        with mock.patch.dict("sys.modules", MySQLdb=m_database):
            call_command("reset_db", "--noinput", verbosity=2)

        m_database.connect.assert_called_once_with(
            host="127.0.0.1", passwd="bar", user="foo"
        )
        m_connection.query.assert_has_calls(expected_calls, any_order=False)
        self.assertEqual("Reset successful.\n", m_stdout.getvalue())

    @override_settings(
        DATABASES={
            "default": {
                "ENGINE": "django.db.backends.mysql",
                "NAME": "test_db",
                "USER": "foo",
                "PASSWORD": "bar",
                "HOST": "/var/run/mysqld/mysql.sock",
                "PORT": "3306",
            },
        }
    )
    @mock.patch("sys.stdout", new_callable=StringIO)
    def test_should_drop_and_create_database_without_characterset_and_print_success_messsage(
        self, m_stdout
    ):
        m_database = mock.MagicMock()
        m_database.__spec__ = mock.Mock()
        m_connection = mock.Mock()
        m_database.connect.return_value = m_connection
        expected_calls = [
            mock.call("DROP DATABASE IF EXISTS `test_db`"),
            mock.call("CREATE DATABASE `test_db`"),
        ]

        with mock.patch.dict("sys.modules", MySQLdb=m_database):
            call_command("reset_db", "--noinput", "--no-utf8", verbosity=2)

        m_database.connect.assert_called_once_with(
            passwd="bar",
            port=3306,
            unix_socket="/var/run/mysqld/mysql.sock",
            user="foo",
        )
        m_connection.query.assert_has_calls(expected_calls, any_order=False)
        self.assertEqual("Reset successful.\n", m_stdout.getvalue())


@override_settings(
    DATABASES={
        "default": {
            "ENGINE": "django.db.backends.postgresql_psycopg2",
            "NAME": "test_db",
            "USER": "foo",
            "PASSWORD": "bar",
            "HOST": "127.0.0.1",
            "PORT": "5432",
        }
    }
)
class ResetDbPostgresqlTests(TestCase):
    """Tests for reset_db command and sqlite3 engine."""

    @mock.patch("sys.stdout", new_callable=StringIO)
    @mock.patch("django_extensions.management.commands.reset_db.input")
    def test_should_cancel_reset_db_if_input_is_different_than_yes(
        self, m_input, m_stdout
    ):
        m_input.return_value = "no"
        call_command("reset_db")
        self.assertEqual("Reset cancelled.\n", m_stdout.getvalue())

    @mock.patch("sys.stdout", new_callable=StringIO)
    def test_should_drop_and_create_database_and_print_success_messsage(self, m_stdout):
        m_database = mock.MagicMock()
        m_database.__spec__ = mock.Mock()
        m_cursor = mock.Mock()
        m_database.connect.return_value.cursor.return_value = m_cursor
        expected_calls = [
            mock.call('DROP DATABASE "test_db";'),
            mock.call(
                'CREATE DATABASE "test_db" WITH OWNER = "foo"  ENCODING = \'UTF8\';'
            ),
        ]

        mock_kwargs = {"psycopg2": m_database}
        has_psycopg3 = importlib.util.find_spec("psycopg") is not None
        if has_psycopg3:
            mock_kwargs = {"psycopg": m_database}
        with mock.patch.dict("sys.modules", **mock_kwargs):
            call_command("reset_db", "--noinput", verbosity=2)

        m_database.connect.assert_called_once_with(
            dbname="template1",
            host="127.0.0.1",
            password="bar",
            port="5432",
            user="foo",
        )

        m_cursor.execute.assert_has_calls(expected_calls, any_order=False)
        self.assertEqual("Reset successful.\n", m_stdout.getvalue())

    @override_settings(DEFAULT_TABLESPACE="TEST_TABLESPACE")
    @mock.patch("sys.stdout", new_callable=StringIO)
    def test_should_drop_create_database_close_sessions_and_print_success_messsage(
        self, m_stdout
    ):
        m_database = mock.MagicMock()
        m_database.__spec__ = mock.Mock()
        m_cursor = mock.Mock()
        m_database.connect.return_value.cursor.return_value = m_cursor
        expected_calls = [
            mock.call(
                "\n                    SELECT pg_terminate_backend(pg_stat_activity.pid)\n                    FROM pg_stat_activity\n                    WHERE pg_stat_activity.datname = 'test_db';\n                "
            ),
            mock.call('DROP DATABASE "test_db";'),
            mock.call(
                'CREATE DATABASE "test_db" WITH OWNER = "foo"  ENCODING = \'UTF8\' TABLESPACE = TEST_TABLESPACE;'
            ),
        ]

        mock_kwargs = {"psycopg2": m_database}
        has_psycopg3 = importlib.util.find_spec("psycopg") is not None
        if has_psycopg3:
            mock_kwargs = {"psycopg": m_database}
        with mock.patch.dict("sys.modules", **mock_kwargs):
            call_command("reset_db", "--noinput", "--close-sessions", verbosity=2)

        m_database.connect.assert_called_once_with(
            dbname="template1",
            host="127.0.0.1",
            password="bar",
            port="5432",
            user="foo",
        )

        m_cursor.execute.assert_has_calls(expected_calls, any_order=False)
        self.assertEqual("Reset successful.\n", m_stdout.getvalue())



# -*- coding: utf-8 -*-
import sys
from django.contrib.auth.models import Group, Permission
from django.test import override_settings

from django_extensions.collision_resolvers import (
    AppNameCR,
    AppsOrderCR,
    BaseCR,
    PathBasedCR,
)
from tests.collisions.models import Group as Group_Col
from tests.collisions.models import Name as Name_Col
from tests.collisions.models import Note as Note_Col
from tests.collisions.models import SystemUser, UniqueModel
from tests.management.commands.shell_plus_tests.test_utils import (
    AutomaticShellPlusImportsTestCase,
)
from tests.testapp.models import Name, Note, UniqueTestAppModel
from tests.testapp.models import Permission as TAPermission


# Bad user defined collision resolvers:


class TestAppsOrderCR(AppsOrderCR):
    pass


class TestAppNameCR(AppNameCR):
    pass


def collision_resolver_which_is_not_class():
    pass


class CRNotExtendingFromBase:
    pass


class CRNoFunction(BaseCR):
    pass


class CRNoArguments(BaseCR):
    def resolve_collisions(self):
        pass


class CRBadResult(BaseCR):
    def resolve_collisions(self, namespace):
        return 1


class CRBadKey(BaseCR):
    def resolve_collisions(self, namespace):
        return {1: "a"}


class CRBadValue(BaseCR):
    def resolve_collisions(self, namespace):
        return {"a": 1}


class CRBadTransformPath(PathBasedCR):
    def transform_import(self, module_path):
        return 1


class CRTestCase(AutomaticShellPlusImportsTestCase):
    def _assert_models_present_under_names(
        self,
        group,
        group_col,
        name,
        name_col,
        note,
        note_col,
        system_user,
        unique_model,
        permission,
        test_app_permission,
        unique_test_app_model,
    ):
        self.run_shell_plus()
        self.assert_imported_under_names(Group, group)
        self.assert_imported_under_names(Group_Col, group_col)
        self.assert_imported_under_names(Name, name)
        self.assert_imported_under_names(Name_Col, name_col)
        self.assert_imported_under_names(Note, note)
        self.assert_imported_under_names(Note_Col, note_col)
        self.assert_imported_under_names(SystemUser, system_user)
        self.assert_imported_under_names(UniqueModel, unique_model)
        self.assert_imported_under_names(Permission, permission)
        self.assert_imported_under_names(TAPermission, test_app_permission)
        self.assert_imported_under_names(UniqueTestAppModel, unique_test_app_model)

    def test_legacy_collision_resolver(self):
        self._assert_models_present_under_names(
            set(),
            {"Group"},
            {"Name"},
            set(),
            {"Note"},
            set(),
            {"SystemUser"},
            {"UniqueModel"},
            set(),
            {"Permission"},
            {"UniqueTestAppModel"},
        )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="django_extensions.collision_resolvers.AppNamePrefixCR",
    )
    def test_app_name_prefix_collision_resolver(self):
        self._assert_models_present_under_names(
            {"django_contrib_auth_Group"},
            {"tests_collisions_Group", "Group"},
            {"django_extensions_Name", "Name"},
            {"tests_collisions_Name"},
            {"django_extensions_Note", "Note"},
            {"tests_collisions_Note"},
            {"SystemUser"},
            {"UniqueModel"},
            {"django_contrib_auth_Permission"},
            {"tests_testapp_Permission", "Permission"},
            {"UniqueTestAppModel"},
        )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="django_extensions.collision_resolvers.AppNameSuffixCR",
    )
    def test_app_name_suffix_collision_resolver(self):
        self._assert_models_present_under_names(
            {"Group_django_contrib_auth"},
            {"Group_tests_collisions", "Group"},
            {"Name_django_extensions", "Name"},
            {"Name_tests_collisions"},
            {"Note_django_extensions", "Note"},
            {"Note_tests_collisions"},
            {"SystemUser"},
            {"UniqueModel"},
            {"Permission_django_contrib_auth"},
            {"Permission_tests_testapp", "Permission"},
            {"UniqueTestAppModel"},
        )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="django_extensions.collision_resolvers.AppLabelPrefixCR",
    )
    def test_app_label_prefix_collision_resolver(self):
        self._assert_models_present_under_names(
            {"auth_Group"},
            {"collisions_Group", "Group"},
            {"django_extensions_Name", "Name"},
            {"collisions_Name"},
            {"django_extensions_Note", "Note"},
            {"collisions_Note"},
            {"SystemUser"},
            {"UniqueModel"},
            {"auth_Permission"},
            {"testapp_Permission", "Permission"},
            {"UniqueTestAppModel"},
        )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="django_extensions.collision_resolvers.AppLabelSuffixCR",
    )
    def test_app_label_suffix_collision_resolver(self):
        self._assert_models_present_under_names(
            {"Group_auth"},
            {"Group_collisions", "Group"},
            {"Name_django_extensions", "Name"},
            {"Name_collisions"},
            {"Note_django_extensions", "Note"},
            {"Note_collisions"},
            {"SystemUser"},
            {"UniqueModel"},
            {"Permission_auth"},
            {"Permission_testapp", "Permission"},
            {"UniqueTestAppModel"},
        )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="django_extensions.collision_resolvers.FullPathCR",
    )
    def test_full_path_collision_resolver(self):
        self._assert_models_present_under_names(
            {"django_contrib_auth_Group"},
            {"tests_collisions_Group", "Group"},
            {"tests_testapp_Name", "Name"},
            {"tests_collisions_Name"},
            {"tests_testapp_Note", "Note"},
            {"tests_collisions_Note"},
            {"SystemUser"},
            {"UniqueModel"},
            {"django_contrib_auth_Permission"},
            {"tests_testapp_Permission", "Permission"},
            {"UniqueTestAppModel"},
        )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="django_extensions.collision_resolvers.InstalledAppsOrderCR",
    )
    def test_installed_apps_order_collision_resolver(self):
        self._assert_models_present_under_names(
            {"Group"},
            set(),
            set(),
            {"Name"},
            set(),
            {"Note"},
            {"SystemUser"},
            {"UniqueModel"},
            {"Permission"},
            set(),
            {"UniqueTestAppModel"},
        )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="tests.management.commands.shell_plus_tests.test_collision_resolver.TestAppsOrderCR",
    )
    def test_installed_bad_order_collision_resolver(self):
        with self.assertRaisesRegex(
            AssertionError, "You must define APP_PRIORITIES in your resolver class!"
        ):
            self._assert_models_present_under_names(
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
            )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="tests.management.commands.shell_plus_tests.test_collision_resolver.TestAppNameCR",
    )
    def test_installed_apps_bad_name_collision_resolver(self):
        with self.assertRaisesRegex(
            AssertionError,
            "You must define MODIFICATION_STRING in your resolver class!",
        ):
            self._assert_models_present_under_names(
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
            )

    def _assert_bad_resolver(self, message):
        with self.assertRaisesRegex(AssertionError, message):
            self._assert_models_present_under_names(
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
            )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="tests.management.commands.shell_plus_tests.test_collision_resolver.collision_resolver_which_is_not_class",
    )
    def test_installed_apps_not_class_collision_resolver(self):
        self._assert_bad_resolver(
            "SHELL_PLUS_MODEL_IMPORTS_RESOLVER must be subclass of BaseCR!"
        )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="tests.management.commands.shell_plus_tests.test_collision_resolver.CRNotExtendingFromBase",
    )
    def test_installed_apps_not_subclass_of_base(self):
        self._assert_bad_resolver(
            "SHELL_PLUS_MODEL_IMPORTS_RESOLVER must be subclass of BaseCR!"
        )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="tests.management.commands.shell_plus_tests.test_collision_resolver.CRNoFunction",
    )
    def test_installed_apps_no_resolve_conflicts_function(self):
        exception_msg = "Can't instantiate abstract class CRNoFunction with abstract methods resolve_collisions"

        if sys.version_info[:2] >= (3, 12):
            exception_msg = "Can't instantiate abstract class CRNoFunction without an implementation for abstract method 'resolve_collisions'"
        elif sys.version_info[:2] >= (3, 9):
            exception_msg = exception_msg.replace("methods", "method")
        with self.assertRaisesRegex(TypeError, exception_msg):
            self._assert_models_present_under_names(
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
                set(),
            )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="tests.management.commands.shell_plus_tests.test_collision_resolver.CRNoArguments",
    )
    def test_installed_apps_no_arguments_resolve_conflicts(self):
        self._assert_bad_resolver("resolve_collisions function must take one argument!")

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="tests.management.commands.shell_plus_tests.test_collision_resolver.CRBadResult",
    )
    def test_installed_apps_bad_result(self):
        self._assert_bad_resolver(
            "Result of resolve_collisions function must be a dict!"
        )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="tests.management.commands.shell_plus_tests.test_collision_resolver.CRBadKey",
    )
    def test_installed_apps_bad_key(self):
        self._assert_bad_resolver(
            "key in collision resolver result should be str not 1"
        )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="tests.management.commands.shell_plus_tests.test_collision_resolver.CRBadValue",
    )
    def test_installed_apps_bad_value(self):
        self._assert_bad_resolver(
            "value in collision resolver result should be str not 1"
        )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="tests.management.commands.shell_plus_tests.test_collision_resolver.CRBadTransformPath"
    )
    def test_bad_transform_path(self):
        self._assert_bad_resolver("result of transform_import must be str!")

    @override_settings(
        SHELL_PLUS_MODEL_ALIASES={"testapp": {"Note": "MyFunnyNote"}},
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="django_extensions.collision_resolvers.AppNamePrefixCR",
    )
    def test_app_name_prefix_collision_resolver_with_model_alias(self):
        self._assert_models_present_under_names(
            {"django_contrib_auth_Group"},
            {"tests_collisions_Group", "Group"},
            {"django_extensions_Name", "Name"},
            {"tests_collisions_Name"},
            {"MyFunnyNote"},
            {"Note"},
            {"SystemUser"},
            {"UniqueModel"},
            {"django_contrib_auth_Permission"},
            {"tests_testapp_Permission", "Permission"},
            {"UniqueTestAppModel"},
        )

    @override_settings(
        SHELL_PLUS_MODEL_ALIASES={"testapp": {"Note": "Name"}},
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="django_extensions.collision_resolvers.AppNamePrefixCR",
    )
    def test_app_name_prefix_collision_resolver_with_clash_because_of_model_alias(self):
        self._assert_models_present_under_names(
            {"django_contrib_auth_Group"},
            {"tests_collisions_Group", "Group"},
            {"django_extensions_Name"},
            {"tests_collisions_Name"},
            {"django_extensions_Note", "Name"},
            {"Note"},
            {"SystemUser"},
            {"UniqueModel"},
            {"django_contrib_auth_Permission"},
            {"tests_testapp_Permission", "Permission"},
            {"UniqueTestAppModel"},
        )

    @override_settings(
        SHELL_PLUS_APP_PREFIXES={"testapp": "main_app"},
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="django_extensions.collision_resolvers.AppNamePrefixCR",
    )
    def test_app_name_prefix_collision_resolver_with_app_prefixes(self):
        self._assert_models_present_under_names(
            {"django_contrib_auth_Group"},
            {"tests_collisions_Group", "Group"},
            {"main_app_Name"},
            {"Name"},
            {"main_app_Note"},
            {"Note"},
            {"SystemUser"},
            {"UniqueModel"},
            {"Permission"},
            {"main_app_Permission"},
            {"main_app_UniqueTestAppModel"},
        )

    @override_settings(
        SHELL_PLUS_APP_PREFIXES={"testapp": "main_app", "collisions": "main_app"},
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="django_extensions.collision_resolvers.AppNamePrefixCR",
    )
    def test_app_name_prefix_collision_resolver_with_clashing_app_prefixes(self):
        self._assert_models_present_under_names(
            {"Group"},
            {"main_app_Group"},
            {"django_extensions_Name", "main_app_Name"},
            {"tests_collisions_Name"},
            {"django_extensions_Note", "main_app_Note"},
            {"tests_collisions_Note"},
            {"main_app_SystemUser"},
            {"main_app_UniqueModel"},
            {"Permission"},
            {"main_app_Permission"},
            {"main_app_UniqueTestAppModel"},
        )

    @override_settings(
        SHELL_PLUS_APP_PREFIXES={"testapp": "main_app", "auth": "main_app"},
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="django_extensions.collision_resolvers.AppNamePrefixCR",
    )
    def test_app_name_prefix_collision_resolver_clash_with_3rd_party(self):
        self._assert_models_present_under_names(
            {"main_app_Group"},
            {"Group"},
            {"main_app_Name"},
            {"Name"},
            {"main_app_Note"},
            {"Note"},
            {"SystemUser"},
            {"UniqueModel"},
            {"django_contrib_auth_Permission"},
            {"main_app_Permission", "tests_testapp_Permission"},
            {"main_app_UniqueTestAppModel"},
        )

    @override_settings(
        SHELL_PLUS_APP_PREFIXES={
            "testapp": "testapp",
            "auth": "auth",
            "collisions": "collisions",
        },
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="django_extensions.collision_resolvers.AppNamePrefixCR",
    )
    def test_no_collisions_because_of_app_prefixes(self):
        self._assert_models_present_under_names(
            {"auth_Group"},
            {"collisions_Group"},
            {"testapp_Name"},
            {"collisions_Name"},
            {"testapp_Note"},
            {"collisions_Note"},
            {"collisions_SystemUser"},
            {"collisions_UniqueModel"},
            {"auth_Permission"},
            {"testapp_Permission"},
            {"testapp_UniqueTestAppModel"},
        )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="django_extensions.collision_resolvers.AppNamePrefixCustomOrderCR",
    )
    def test_app_name_prefix_custom_order_collision_resolver(self):
        self._assert_models_present_under_names(
            {"django_contrib_auth_Group", "Group"},
            {"tests_collisions_Group"},
            {"django_extensions_Name"},
            {"tests_collisions_Name", "Name"},
            {"django_extensions_Note"},
            {"tests_collisions_Note", "Note"},
            {"SystemUser"},
            {"UniqueModel"},
            {"django_contrib_auth_Permission", "Permission"},
            {"tests_testapp_Permission"},
            {"UniqueTestAppModel"},
        )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="django_extensions.collision_resolvers.AppNameSuffixCustomOrderCR",
    )
    def test_app_name_suffix_custom_order_collision_resolver(self):
        self._assert_models_present_under_names(
            {"Group_django_contrib_auth", "Group"},
            {"Group_tests_collisions"},
            {"Name_django_extensions"},
            {"Name_tests_collisions", "Name"},
            {"Note_django_extensions"},
            {"Note_tests_collisions", "Note"},
            {"SystemUser"},
            {"UniqueModel"},
            {"Permission_django_contrib_auth", "Permission"},
            {"Permission_tests_testapp"},
            {"UniqueTestAppModel"},
        )

    @override_settings(
        SHELL_PLUS_MODEL_IMPORTS_RESOLVER="django_extensions.collision_resolvers.FullPathCustomOrderCR",
    )
    def test_full_path_custom_order_collision_resolver(self):
        self._assert_models_present_under_names(
            {"django_contrib_auth_Group", "Group"},
            {"tests_collisions_Group"},
            {"tests_testapp_Name"},
            {"tests_collisions_Name", "Name"},
            {"tests_testapp_Note"},
            {"tests_collisions_Note", "Note"},
            {"SystemUser"},
            {"UniqueModel"},
            {"django_contrib_auth_Permission", "Permission"},
            {"tests_testapp_Permission"},
            {"UniqueTestAppModel"},
        )



# -*- coding: utf-8 -*-
import importlib.util
from io import StringIO
from unittest.mock import MagicMock, Mock, PropertyMock, call, patch

from django.core.management import CommandError, call_command
from django.test import TestCase
from django.test.utils import override_settings


# Database testing configurations

UNKOWN_ENGINE = {
    "default": {
        "ENGINE": "django.db.backends.unknown",
        "NAME": "unknown",
    }
}

NO_TEST_NAME = {
    "default": {
        "ENGINE": "django.db.backends.mysql",
        "NAME": "test",
        "TEST": {
            "NAME": "",
        },
    }
}

SQLITE = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "db.sqlite3",
    }
}

MYSQL_HOST_PORT = {
    "default": {
        "ENGINE": "django.db.backends.mysql",
        "NAME": "test",
        "USER": "test",
        "PASSWORD": "test",
        "HOST": "localhost",
        "PORT": "3306",
    },
}

MYSQL_SOCKET = {
    "default": {
        "ENGINE": "django.db.backends.mysql",
        "NAME": "test",
        "USER": "test",
        "PASSWORD": "test",
        "HOST": "/var/run/mysqld/mysql.sock",
    },
}

POSTGRES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql_psycopg2",
        "NAME": "test",
        "USER": "test",
        "PASSWORD": "test",
        "PORT": "5432",
        "HOST": "localhost",
    },
}


class DropTestDatabaseExceptionsTests(TestCase):
    """Test for drop_test_database command."""

    def test_should_raise_CommandError_if_database_is_unknown(self):
        with self.assertRaisesRegex(CommandError, "Unknown database unknown"):
            call_command("drop_test_database", "--database=unknown")

    @override_settings(DATABASES=UNKOWN_ENGINE)
    @patch("django_extensions.management.commands.drop_test_database.input")
    def test_should_raise_CommandError_if_unknown_database_engine(self, m_input):
        m_input.return_value = "yes"
        with self.assertRaisesRegex(
            CommandError, "Unknown database engine django.db.backends.unknown"
        ):
            call_command("drop_test_database")

    @override_settings(DATABASES=NO_TEST_NAME)
    def test_should_raise_CommandError_if_test_database_name_is_empty(self):
        with self.assertRaisesRegex(
            CommandError,
            "You need to specify DATABASE_NAME in your Django settings file.",
        ):
            call_command("drop_test_database")


class DropTestDatabaseTests(TestCase):
    """Test for drop_test_database command."""

    @patch("sys.stdout", new_callable=StringIO)
    @patch("django_extensions.management.commands.drop_test_database.input")
    def test_should_raise_CommandError_if_database_is_unknown(self, m_input, m_stdout):
        m_input.return_value = "no"

        call_command("drop_test_database")

        self.assertEqual("Reset cancelled.\n", m_stdout.getvalue())

    @override_settings(DATABASES=SQLITE)
    @patch("sys.stdout", new_callable=StringIO)
    @patch("os.path.isfile")
    @patch("os.unlink")
    def test_sqlite3_should_unlink_primary_test_database(
        self, m_unlink, m_isfile, m_stdout
    ):
        # Indicate that no clone databases exist
        m_isfile.side_effect = (True, False)
        call_command("drop_test_database", "--noinput", verbosity=2)

        with self.subTest("Should check for test database names until failure"):
            self.assertListEqual(
                m_isfile.call_args_list,
                # See production code comments regarding double dots
                [call("test_db.sqlite3"), call("test_db_1..sqlite3")],
            )

        with self.subTest("Should unlink only primary test database"):
            self.assertListEqual(
                m_unlink.call_args_list,
                [call("test_db.sqlite3")],
            )

        with self.subTest("Should report successful message"):
            self.assertIn("Reset successful.", m_stdout.getvalue())

    @override_settings(DATABASES=SQLITE)
    @patch("os.path.isfile")
    @patch("os.unlink")
    def test_sqlite3_should_unlink_all_existing_clone_databases(
        self, m_unlink, m_isfile
    ):
        """Test cloned test databases created via 'manage.py test --parallel'."""
        # Indicate that clone databases exist up to test_db_2.sqlite3
        m_isfile.side_effect = (True, True, True, False)
        call_command("drop_test_database", "--noinput")

        with self.subTest("Should check for test database names until failure"):
            self.assertListEqual(
                m_isfile.call_args_list,
                [
                    call("test_db.sqlite3"),
                    # See production code comments regarding double dots
                    call("test_db_1..sqlite3"),
                    call("test_db_2..sqlite3"),
                    call("test_db_3..sqlite3"),
                ],
            )

        with self.subTest("Should unlink all existing test databases"):
            self.assertListEqual(
                m_unlink.call_args_list,
                [
                    call("test_db.sqlite3"),
                    # See production code comments regarding double dots
                    call("test_db_1..sqlite3"),
                    call("test_db_2..sqlite3"),
                ],
            )

    @override_settings(DATABASES=SQLITE)
    @patch("sys.stdout", new_callable=StringIO)
    @patch("os.path.isfile")
    @patch("os.unlink")
    def test_sqlite3_should_not_print_Reset_successful_when_OSError_exception(
        self, m_unlink, m_isfile, m_stdout
    ):
        m_isfile.return_value = True
        m_unlink.side_effect = OSError
        call_command("drop_test_database", "--noinput", verbosity=2)

        self.assertNotIn("Reset successful.", m_stdout.getvalue())

    @override_settings(DATABASES=MYSQL_HOST_PORT)
    @patch("sys.stdout", new_callable=StringIO)
    def test_mysql_should_drop_database_with_host_and_port(self, m_stdout):
        m_database = MagicMock()
        m_database.__spec__ = Mock()
        # Indicate that no clone databases exist
        # DROP queries return None while SELECT queries return a row count
        m_database.connect.return_value.cursor.return_value.execute.side_effect = (
            1,
            None,
            0,
        )

        with patch.dict("sys.modules", MySQLdb=m_database):
            call_command("drop_test_database", "--noinput", verbosity=2)

        with self.subTest(
            "Should check for and remove test database names until failure"
        ):
            exists_query = (
                "SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME="
            )
            self.assertListEqual(
                m_database.connect.return_value.cursor.return_value.execute.call_args_list,
                [
                    call(exists_query + "'test_test';"),
                    call("DROP DATABASE IF EXISTS `test_test`"),
                    call(exists_query + "'test_test_1';"),
                ],
            )

        with self.subTest("Should report successful message"):
            self.assertIn("Reset successful.", m_stdout.getvalue())

    @override_settings(DATABASES=MYSQL_SOCKET)
    @patch("sys.stdout", new_callable=StringIO)
    def test_mysql_should_drop_database_with_unix_socket(self, m_stdout):
        m_database = MagicMock()
        m_database.__spec__ = Mock()
        # Indicate that no clone databases exist
        # DROP queries return None while SELECT queries return a row count
        m_database.connect.return_value.cursor.return_value.execute.side_effect = (
            1,
            None,
            0,
        )

        with patch.dict("sys.modules", MySQLdb=m_database):
            call_command("drop_test_database", "--noinput", verbosity=2)

        with self.subTest(
            "Should check for and remove test database names until failure"
        ):
            exists_query = (
                "SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME="
            )
            self.assertListEqual(
                m_database.connect.return_value.cursor.return_value.execute.call_args_list,
                [
                    call(exists_query + "'test_test';"),
                    call("DROP DATABASE IF EXISTS `test_test`"),
                    call(exists_query + "'test_test_1';"),
                ],
            )

        with self.subTest("Should report successful message"):
            self.assertIn("Reset successful.", m_stdout.getvalue())

    @override_settings(DATABASES=MYSQL_HOST_PORT)
    def test_mysql_should_drop_all_existing_clone_databases(self):
        """Test cloned test databases created via 'manage.py test --parallel'."""
        m_database = MagicMock()
        m_database.__spec__ = Mock()
        # Indicate that clone databases exist up to test_test_2
        # DROP queries return None while SELECT queries return a row count
        m_database.connect.return_value.cursor.return_value.execute.side_effect = (
            1,
            None,
            1,
            None,
            1,
            None,
            0,
        )

        with patch.dict("sys.modules", MySQLdb=m_database):
            call_command("drop_test_database", "--noinput")

        exists_query = (
            "SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME="
        )
        self.assertListEqual(
            m_database.connect.return_value.cursor.return_value.execute.call_args_list,
            [
                call(exists_query + "'test_test';"),
                call("DROP DATABASE IF EXISTS `test_test`"),
                call(exists_query + "'test_test_1';"),
                call("DROP DATABASE IF EXISTS `test_test_1`"),
                call(exists_query + "'test_test_2';"),
                call("DROP DATABASE IF EXISTS `test_test_2`"),
                call(exists_query + "'test_test_3';"),
            ],
        )

    @override_settings(DATABASES=POSTGRES)
    @patch("sys.stdout", new_callable=StringIO)
    def test_postgresql_should_drop_database(self, m_stdout):
        m_database = MagicMock()
        m_database.__spec__ = Mock()
        m_cursor = Mock()
        m_database.connect.return_value.cursor.return_value = m_cursor
        # Indicate that no clone databases exist
        type(m_cursor).rowcount = PropertyMock(side_effect=(1, 0))

        mock_kwargs = {"psycopg2": m_database}
        has_psycopg3 = importlib.util.find_spec("psycopg") is not None
        if has_psycopg3:
            mock_kwargs = {"psycopg": m_database}
        with patch.dict("sys.modules", **mock_kwargs):
            call_command("drop_test_database", "--noinput", verbosity=2)

        with self.subTest(
            "Should check for and remove test database names until failure"
        ):
            exists_query = "SELECT datname FROM pg_catalog.pg_database WHERE datname="
            self.assertListEqual(
                m_cursor.execute.call_args_list,
                [
                    call(exists_query + "'test_test';"),
                    call('DROP DATABASE IF EXISTS "test_test";'),
                    call(exists_query + "'test_test_1';"),
                ],
            )

        with self.subTest("Should report successful message"):
            self.assertIn("Reset successful.", m_stdout.getvalue())

    @override_settings(DATABASES=POSTGRES)
    def test_postgresql_should_drop_all_existing_cloned_databases(self):
        """Test cloned test databases created via 'manage.py test --parallel'."""
        m_database = MagicMock()
        m_database.__spec__ = Mock()
        m_cursor = Mock()
        m_database.connect.return_value.cursor.return_value = m_cursor
        # Indicate that clone databases exist up to test_test_2
        type(m_cursor).rowcount = PropertyMock(side_effect=(1, 1, 1, 0))

        mock_kwargs = {"psycopg2": m_database}
        has_psycopg3 = importlib.util.find_spec("psycopg") is not None
        if has_psycopg3:
            mock_kwargs = {"psycopg": m_database}
        with patch.dict("sys.modules", **mock_kwargs):
            call_command("drop_test_database", "--noinput")

        exists_query = "SELECT datname FROM pg_catalog.pg_database WHERE datname="
        self.assertListEqual(
            m_cursor.execute.call_args_list,
            [
                call(exists_query + "'test_test';"),
                call('DROP DATABASE IF EXISTS "test_test";'),
                call(exists_query + "'test_test_1';"),
                call('DROP DATABASE IF EXISTS "test_test_1";'),
                call(exists_query + "'test_test_2';"),
                call('DROP DATABASE IF EXISTS "test_test_2";'),
                call(exists_query + "'test_test_3';"),
            ],
        )

    @override_settings(DATABASES=POSTGRES)
    @patch("sys.stdout", new_callable=StringIO)
    def test_postgresql_should_not_print_Reset_successful_when_exception_occured(
        self, m_stdout
    ):
        m_database = MagicMock()
        m_database.__spec__ = Mock()
        m_database.ProgrammingError = Exception
        m_cursor = Mock()
        m_cursor.execute.side_effect = m_database.ProgrammingError
        m_database.connect.return_value.cursor.return_value = m_cursor

        mock_kwargs = {"psycopg2": m_database}
        has_psycopg3 = importlib.util.find_spec("psycopg") is not None
        if has_psycopg3:
            mock_kwargs = {"psycopg": m_database}
        with patch.dict("sys.modules", **mock_kwargs):
            call_command("drop_test_database", "--noinput", verbosity=2)

        self.assertNotIn("Reset successful.", m_stdout.getvalue())



# -*- coding: utf-8 -*-
import os
import sys
import shutil
from typing import List

from django.core.management.base import AppCommand
from django.core.management.color import color_style

from django_extensions.management.utils import _make_writeable, signalcommand


class Command(AppCommand):
    help = "Creates a Django management command directory structure for the given app "
    "name in the app's directory."

    requires_system_checks: List[str] = []
    # Can't import settings during this command, because they haven't
    # necessarily been created.
    can_import_settings = True

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--name",
            "-n",
            action="store",
            dest="command_name",
            default="sample",
            help="The name to use for the management command",
        )
        parser.add_argument(
            "--base",
            "-b",
            action="store",
            dest="base_command",
            default="Base",
            help="The base class used for implementation of "
            "this command. Should be one of Base, App, Label, or NoArgs",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            default=False,
            help="Do not actually create any files",
        )

    @signalcommand
    def handle_app_config(self, args, **options):
        app = args
        copy_template("command_template", app.path, **options)


def copy_template(template_name, copy_to, **options):
    """Copy the specified template directory to the copy_to location"""
    import django_extensions

    style = color_style()
    ERROR = getattr(style, "ERROR", lambda x: x)
    SUCCESS = getattr(style, "SUCCESS", lambda x: x)

    command_name, base_command = (
        options["command_name"],
        "%sCommand" % options["base_command"],
    )
    dry_run = options["dry_run"]
    verbosity = options["verbosity"]

    template_dir = os.path.join(django_extensions.__path__[0], "conf", template_name)

    # walk the template structure and copies it
    for d, subdirs, files in os.walk(template_dir):
        relative_dir = d[len(template_dir) + 1 :]
        if relative_dir and not os.path.exists(os.path.join(copy_to, relative_dir)):
            if not dry_run:
                os.mkdir(os.path.join(copy_to, relative_dir))
        for i, subdir in enumerate(subdirs):
            if subdir.startswith("."):
                del subdirs[i]
        for f in files:
            if f.endswith((".pyc", ".pyo")) or f.startswith(
                (".DS_Store", "__pycache__")
            ):
                continue
            path_old = os.path.join(d, f)
            path_new = os.path.join(
                copy_to, relative_dir, f.replace("sample", command_name)
            ).rstrip(".tmpl")
            if os.path.exists(path_new):
                path_new = os.path.join(copy_to, relative_dir, f).rstrip(".tmpl")
                if os.path.exists(path_new):
                    if verbosity > 1:
                        print(ERROR("%s already exists" % path_new))
                    continue
            if verbosity > 1:
                print(SUCCESS("%s" % path_new))
            with open(path_old, "r") as fp_orig:
                data = fp_orig.read()
                data = data.replace("{{ command_name }}", command_name)
                data = data.replace("{{ base_command }}", base_command)
                if not dry_run:
                    with open(path_new, "w") as fp_new:
                        fp_new.write(data)
            if not dry_run:
                try:
                    shutil.copymode(path_old, path_new)
                    _make_writeable(path_new)
                except OSError:
                    sys.stderr.write(
                        "Notice: Couldn't set permission bits on %s. You're probably using an uncommon filesystem setup. No problem.\n"  # noqa: E501
                        % path_new
                    )



# -*- coding: utf-8 -*-
from django.db import models


class Note(models.Model):
    # name conflict with testapp.Note
    text = models.CharField(max_length=42)
    creation_date = models.DateField(auto_now_add=True)

    class Meta:
        app_label = "collisions"


class UniqueModel(models.Model):
    # no conflicts
    global_hash_id = models.CharField(max_length=32)

    class Meta:
        app_label = "collisions"


class Group(models.Model):
    # conflict with django.contrib.auth
    name = models.CharField(max_length=10)
    priority = models.IntegerField()


class Name(models.Model):
    # name conflict with testapp.Name
    real_name = models.CharField(max_length=50)
    number_of_users_having_this_name = models.IntegerField()

    class Meta:
        app_label = "collisions"


class SystemUser(models.Model):
    # no conflicts but FK to conflicting models.
    name = models.ForeignKey(Name, on_delete=models.CASCADE)
    group = models.ForeignKey(Group, on_delete=models.CASCADE)
    global_id = models.CharField(unique=True, max_length=32)



# -*- coding: utf-8 -*-
"""
Similar to syntax_color.py but this is intended more for being able to
copy+paste actual code into your Django templates without needing to
escape or anything crazy.

Example:

 {% load highlighting %}

 <style>
 .highlight { background: #f8f8f8; }
 .highlight { font-size: 11px; margin: 1em; border: 1px solid #ccc;
              border-left: 3px solid #F90; padding: 0; }
 .highlight pre { padding: 1em; overflow: auto; line-height: 120%; margin: 0; }
 .predesc { margin: 1.5em 1.5em -2.5em 1em; text-align: right;
            font: bold 12px Tahoma, Arial, sans-serif;
            letter-spacing: 1px; color: #333; }
 </style>

 <h2>check out this code</h2>

 {% highlight 'python' 'Excerpt: blah.py' %}
 def need_food(self):
     print("Love is <colder> than &death&")
 {% endhighlight %}
"""

from django import template
from django.template import (
    Context,
    Node,
    Template,
    TemplateSyntaxError,
    Variable,
)
from django.template.defaultfilters import stringfilter
from django.utils.safestring import mark_safe

try:
    from pygments import highlight as pyghighlight
    from pygments.lexers import get_lexer_by_name
    from pygments.formatters import HtmlFormatter

    HAS_PYGMENTS = True
except ImportError:  # pragma: no cover
    HAS_PYGMENTS = False

register = template.Library()


@register.filter(is_safe=True)
@stringfilter
def parse_template(value):
    return mark_safe(Template(value).render(Context()))


class CodeNode(Node):
    def __init__(self, language, nodelist, name=""):
        self.language = Variable(language)
        self.nodelist = nodelist
        if name:
            self.name = Variable(name)
        else:
            self.name = None

    def render(self, context):
        code = self.nodelist.render(context).strip()
        lexer = get_lexer_by_name(self.language.resolve(context))
        formatter = HtmlFormatter(linenos=False)
        html = ""
        if self.name:
            name = self.name.resolve(context)
            html = '<div class="predesc"><span>%s</span></div>' % name
        return html + pyghighlight(code, lexer, formatter)


@register.tag
def highlight(parser, token):
    """
    Tag to put a highlighted source code <pre> block in your code.
    This takes two arguments, the language and a little explaination message
    that will be generated before the code.  The second argument is optional.

    Your code will be fed through pygments so you can use any language it
    supports.

    Usage::

      {% load highlighting %}
      {% highlight 'python' 'Excerpt: blah.py' %}
      def need_food(self):
          print("Love is colder than death")
      {% endhighlight %}

    """
    if not HAS_PYGMENTS:  # pragma: no cover
        raise ImportError("Please install 'pygments' library to use highlighting.")
    nodelist = parser.parse(("endhighlight",))
    parser.delete_first_token()
    bits = token.split_contents()[1:]
    if len(bits) < 1:
        raise TemplateSyntaxError("'highlight' statement requires an argument")
    return CodeNode(bits[0], nodelist, *bits[1:])



# -*- coding: utf-8 -*-
from django.core.cache import caches
from django.core.management import call_command
from django.test import TestCase
from django.test.utils import override_settings


class CacheCleanupTests(TestCase):
    """Tests for cache_cleanup job."""

    def test_should_not_do_anything_if_settings_does_not_have_CACHES_settings(self):
        call_command("runjob", "cache_cleanup", verbosity=2)

    @override_settings(
        CACHES={
            "test_cache": {
                "BACKEND": "django.core.cache.backends.db.DatabaseCache",
                "LOCATION": "my_cache_table",
            }
        }
    )
    def test_remove_all_keys_from_DatabaseCache(self):
        call_command("createcachetable")
        db_cache = caches["test_cache"]
        db_cache.set("my_key", "hello world")

        call_command("runjob", "cache_cleanup", verbosity=2)

        self.assertIsNone(db_cache.get("my_key"))



# -*- coding: utf-8 -*-
import os
import shutil
from io import StringIO
from tempfile import mkdtemp

from django.conf import settings
from django.core.management import CommandError, call_command
from django.test import TestCase
from django.test.utils import override_settings


def test_validate_templates():
    out = StringIO()
    try:
        call_command("validate_templates", verbosity=3, stdout=out, stderr=out)
    except CommandError:
        print(out.getvalue())
        raise

    output = out.getvalue()
    assert "0 errors found\n" in output


class ValidateTemplatesTests(TestCase):
    def setUp(self):
        self.tempdir = mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_should_print_that_there_is_error(self):
        with override_settings(
            INSTALLED_APPS=settings.INSTALLED_APPS
            + ["tests.testapp_with_template_errors"]
        ):
            with self.assertRaisesRegex(CommandError, "1 errors found"):
                call_command("validate_templates", verbosity=3)

    def test_should_not_print_any_errors_if_template_in_VALIDATE_TEMPLATES_IGNORES(
        self,
    ):
        out = StringIO()

        with override_settings(
            INSTALLED_APPS=settings.INSTALLED_APPS
            + ["tests.testapp_with_template_errors"],
            VALIDATE_TEMPLATES_IGNORES=["template_with_error.html"],
        ):
            call_command("validate_templates", verbosity=3, stdout=out, stderr=out)

        output = out.getvalue()

        self.assertIn("0 errors found\n", output)

    def test_should_not_print_any_errors_if_app_in_VALIDATE_TEMPLATES_IGNORE_APPS(self):
        out = StringIO()

        with override_settings(
            INSTALLED_APPS=settings.INSTALLED_APPS
            + ["tests.testapp_with_template_errors"],
            VALIDATE_TEMPLATES_IGNORE_APPS=["tests.testapp_with_template_errors"],
        ):
            call_command("validate_templates", verbosity=3, stdout=out, stderr=out)

        output = out.getvalue()

        self.assertIn("0 errors found\n", output)

    def test_should_break_when_first_error_occur(self):
        fn = os.path.join(self.tempdir, "template_with_error.html")
        with open(fn, "w") as f:
            f.write("""{% invalid_tag %}""")

        with self.assertRaisesRegex(CommandError, "Errors found"):
            call_command("validate_templates", "-i", self.tempdir, "-b", verbosity=3)



# -*- coding: utf-8 -*-
from io import StringIO

from django.core.management import call_command
from django.db import models
from django.test import TestCase


class AdminGeneratorTests(TestCase):
    """Tests for admin_generator management command."""

    def setUp(self):
        self.out = StringIO()

    def test_command(self):
        call_command("admin_generator", "django_extensions", stdout=self.out)
        output = self.out.getvalue()
        self.assertIn("@admin.register(Secret)", output)
        self.assertIn("class SecretAdmin(admin.ModelAdmin):", output)
        self.assertIn("list_display = ('id', 'name', 'text')", output)
        self.assertIn("search_fields = ('name',)", output)

    def test_should_print_warning_if_given_app_is_not_installed(self):
        expected_output = """This command requires an existing app name as argument
Available apps:
    admin"""
        call_command("admin_generator", "invalid_app", stderr=self.out)

        for expected_line in expected_output.splitlines():
            self.assertIn(expected_line, self.out.getvalue())

    def test_should_print_admin_class_for_User_model_only(self):
        call_command("admin_generator", "auth", "Group", stdout=self.out)

        self.assertIn("from .models import Group", self.out.getvalue())

    def test_should_print_admin_class_with_date_hierarchy(self):
        class TestAdminModel(models.Model):
            created_at = models.DateTimeField(auto_now_add=True)
            title = models.CharField(max_length=50)

            class Meta:
                app_label = "testapp"

        call_command("admin_generator", "testapp", "TestAdminModel", stdout=self.out)

        self.assertIn("date_hierarchy = 'created_at'", self.out.getvalue())



# -*- coding: utf-8 -*-
from io import StringIO

from django.test import TestCase
from django.db import models
from django.core.management import CommandError, call_command


class DescribeFormExceptionsTests(TestCase):
    """Tests for describe_form command exceptions."""

    def test_should_raise_CommandError_if_invalid_arg(self):
        with self.assertRaisesRegex(
            CommandError, "Need application and model name in the form: appname.model"
        ):
            call_command("describe_form", "testapp")


class DescribeFormTests(TestCase):
    """Tests for describe_form command."""

    def setUp(self):
        self.out = StringIO()

        class BaseModel(models.Model):
            title = models.CharField(max_length=50)
            body = models.TextField()

            class Meta:
                app_label = "testapp"

        class NonEditableModel(models.Model):
            created_at = models.DateTimeField(auto_now_add=True)
            title = models.CharField(max_length=50)

            class Meta:
                app_label = "testapp"

    def test_should_print_form_definition_for_TestModel(self):
        expected_result = """from django import forms
from testapp.models import BaseModel

class BaseModelForm(forms.Form):
    title = forms.CharField(label='Title', max_length=50)
    body = forms.CharField(label='Body')"""

        call_command("describe_form", "testapp.BaseModel", stdout=self.out)

        self.assertIn(expected_result, self.out.getvalue())

    def test_should_print_form_definition_for_TestModel_with_non_editable_field(self):
        expected_result = """from django import forms
from testapp.models import NonEditableModel

class NonEditableModelForm(forms.Form):
    title = forms.CharField(label='Title', max_length=50)"""

        call_command("describe_form", "testapp.NonEditableModel", stdout=self.out)

        self.assertIn(expected_result, self.out.getvalue())

    def test_should_print_form_with_fields_for_TestModel(self):
        not_expected = """body = forms.CharField(label='Body')"""

        call_command(
            "describe_form", "testapp.BaseModel", "--fields=title", stdout=self.out
        )

        self.assertNotIn(not_expected, self.out.getvalue())



# -*- coding: utf-8 -*-
from django.template import engines
from django.test import TestCase

from django_extensions.templatetags.widont import widont, widont_html

from unittest.mock import MagicMock


# TODO: these tests are far from having decent test coverage
class TemplateTagsTests(TestCase):
    def test_widont(self):
        self.assertEqual(widont("Test Value"), "Test&nbsp;Value")
        self.assertEqual(widont(str("Test Value")), str("Test&nbsp;Value"))

    def test_widont_html(self):
        self.assertEqual(widont_html("Test Value"), "Test&nbsp;Value")
        self.assertEqual(widont_html(str("Test Value")), str("Test&nbsp;Value"))


class DebuggerTagsTests(TestCase):
    """Test class for DebuggerTags."""

    def setUp(self):
        self.engine = engines["django"]

    def test_pdb_filter(self):
        """Test for pdb filter."""
        import pdb

        pdb.set_trace = MagicMock(return_value=None)
        template = self.engine.from_string(
            """
            {% load debugger_tags %}

            {{ test_object|pdb }}
            """
        )
        template.render({"test_object": "test_value"})
        self.assertTrue(pdb.set_trace.called)



# -*- coding: utf-8 -*-
import urllib

from django import forms
from django.contrib.admin.sites import site
from django.contrib.admin.widgets import ForeignKeyRawIdWidget
from django.template.loader import render_to_string
from django.templatetags.static import static
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.text import Truncator


class ForeignKeySearchInput(ForeignKeyRawIdWidget):
    """
    Widget for displaying ForeignKeys in an autocomplete search input
    instead in a <select> box.
    """

    # Set in subclass to render the widget with a different template
    widget_template = None
    # Set this to the patch of the search view
    search_path = None

    @property
    def media(self):
        js_files = [
            static("django_extensions/js/jquery.bgiframe.js"),
            static("django_extensions/js/jquery.ajaxQueue.js"),
            static("django_extensions/js/jquery.autocomplete.js"),
        ]

        return forms.Media(
            css={"all": (static("django_extensions/css/jquery.autocomplete.css"),)},
            js=js_files,
        )

    def label_for_value(self, value):
        key = self.rel.get_related_field().name
        obj = self.rel.model._default_manager.get(**{key: value})

        return Truncator(obj).words(14, truncate="...")

    def __init__(self, rel, search_fields, attrs=None):
        self.search_fields = search_fields
        super().__init__(rel, site, attrs)

    def render(self, name, value, attrs=None, renderer=None):
        if attrs is None:
            attrs = {}
        opts = self.rel.model._meta
        app_label = opts.app_label
        model_name = opts.object_name.lower()
        related_url = reverse("admin:%s_%s_changelist" % (app_label, model_name))
        if not self.search_path:
            self.search_path = urllib.parse.urljoin(
                related_url, "foreignkey_autocomplete/"
            )
        params = self.url_parameters()
        if params:
            url = "?" + "&amp;".join(["%s=%s" % (k, v) for k, v in params.items()])
        else:
            url = ""

        if "class" not in attrs:
            attrs["class"] = "vForeignKeyRawIdAdminField"
        # Call the TextInput render method directly to have more control
        output = [forms.TextInput.render(self, name, value, attrs)]

        if value:
            label = self.label_for_value(value)
        else:
            label = ""

        context = {
            "url": url,
            "related_url": related_url,
            "search_path": self.search_path,
            "search_fields": ",".join(self.search_fields),
            "app_label": app_label,
            "model_name": model_name,
            "label": label,
            "name": name,
        }
        output.append(
            render_to_string(
                self.widget_template
                or (
                    "django_extensions/widgets/%s/%s/foreignkey_searchinput.html"
                    % (app_label, model_name),
                    "django_extensions/widgets/%s/foreignkey_searchinput.html"
                    % app_label,
                    "django_extensions/widgets/foreignkey_searchinput.html",
                ),
                context,
            )
        )
        output.reverse()

        return mark_safe("".join(output))



# -*- coding: utf-8 -*-
"""
      Title: Dumpscript management command
    Project: Hardytools (queryset-refactor version)
     Author: Will Hardy
       Date: June 2008
      Usage: python manage.py dumpscript appname > scripts/scriptname.py
  $Revision: 217 $

Description:
    Generates a Python script that will repopulate the database using objects.
    The advantage of this approach is that it is easy to understand, and more
    flexible than directly populating the database, or using XML.

    * It also allows for new defaults to take effect and only transfers what is
      needed.
    * If a new database schema has a NEW ATTRIBUTE, it is simply not
      populated (using a default value will make the transition smooth :)
    * If a new database schema REMOVES AN ATTRIBUTE, it is simply ignored
      and the data moves across safely (I'm assuming we don't want this
      attribute anymore.
    * Problems may only occur if there is a new model and is now a required
      ForeignKey for an existing model. But this is easy to fix by editing the
      populate script. Half of the job is already done as all ForeingKey
      lookups occur though the locate_object() function in the generated script.

Improvements:
    See TODOs and FIXMEs scattered throughout :-)

"""

import datetime
import sys

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand
from django.db import router
from django.db.models import (
    AutoField,
    BooleanField,
    DateField,
    DateTimeField,
    FileField,
    ForeignKey,
)
from django.db.models.deletion import Collector
from django.utils import timezone
from django.utils.encoding import force_str, smart_str

from django_extensions.management.utils import signalcommand


def orm_item_locator(orm_obj):
    """
    Is called every time an object that will not be exported is required.
    Where orm_obj is the referred object.
    We postpone the lookup to locate_object() which will be run on the generated script
    """

    the_class = orm_obj._meta.object_name
    original_class = the_class
    pk_name = orm_obj._meta.pk.name
    original_pk_name = pk_name
    pk_value = getattr(orm_obj, pk_name)

    while (
        hasattr(pk_value, "_meta")
        and hasattr(pk_value._meta, "pk")
        and hasattr(pk_value._meta.pk, "name")
    ):
        the_class = pk_value._meta.object_name
        pk_name = pk_value._meta.pk.name
        pk_value = getattr(pk_value, pk_name)

    clean_dict = make_clean_dict(orm_obj.__dict__)

    for key in clean_dict:
        v = clean_dict[key]
        if v is not None:
            if isinstance(v, datetime.datetime):
                if not timezone.is_aware(v):
                    v = timezone.make_aware(v)
                clean_dict[key] = StrToCodeChanger(
                    'dateutil.parser.parse("%s")' % v.isoformat()
                )
            elif not isinstance(v, (str, int, float)):
                clean_dict[key] = str("%s" % v)

    output = """ importer.locate_object(%s, "%s", %s, "%s", %s, %s ) """ % (
        original_class,
        original_pk_name,
        the_class,
        pk_name,
        pk_value,
        clean_dict,
    )

    return output


class Command(BaseCommand):
    help = "Dumps the data as a customised python script."

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument("appname", nargs="+")
        parser.add_argument(
            "--autofield",
            action="store_false",
            dest="skip_autofield",
            default=True,
            help="Include Autofields (like pk fields)",
        )

    @signalcommand
    def handle(self, *args, **options):
        app_labels = options["appname"]

        # Get the models we want to export
        models = get_models(app_labels)

        # A dictionary is created to keep track of all the processed objects,
        # so that foreign key references can be made using python variable names.
        # This variable "context" will be passed around like the town bicycle.
        context = {}

        # Create a dumpscript object and let it format itself as a string
        script = Script(
            models=models,
            context=context,
            stdout=self.stdout,
            stderr=self.stderr,
            options=options,
        )
        self.stdout.write(str(script))
        self.stdout.write("\n")


def get_models(app_labels):
    """
    Get a list of models for the given app labels, with some exceptions.
    TODO: If a required model is referenced, it should also be included.
    Or at least discovered with a get_or_create() call.
    """

    # These models are not to be outputted,
    # e.g. because they can be generated automatically
    # TODO: This should be "appname.modelname" string
    EXCLUDED_MODELS = (ContentType,)

    models = []

    # If no app labels are given, return all
    if not app_labels:
        for app in apps.get_app_configs():
            models += [
                m
                for m in apps.get_app_config(app.label).get_models()
                if m not in EXCLUDED_MODELS
            ]
        return models

    # Get all relevant apps
    for app_label in app_labels:
        # If a specific model is mentioned, get only that model
        if "." in app_label:
            app_label, model_name = app_label.split(".", 1)
            models.append(apps.get_model(app_label, model_name))
        # Get all models for a given app
        else:
            models += [
                m
                for m in apps.get_app_config(app_label).get_models()
                if m not in EXCLUDED_MODELS
            ]

    return models


class Code:
    """
    A snippet of python script.
    This keeps track of import statements and can be output to a string.
    In the future, other features such as custom indentation might be included
    in this class.
    """

    def __init__(self, indent=-1, stdout=None, stderr=None):
        if not stdout:
            stdout = sys.stdout
        if not stderr:
            stderr = sys.stderr

        self.indent = indent
        self.stdout = stdout
        self.stderr = stderr

    def __str__(self):
        """Return a string representation of this script."""
        if self.imports:
            self.stderr.write(repr(self.import_lines))
            return flatten_blocks(
                [""] + self.import_lines + [""] + self.lines, num_indents=self.indent
            )
        else:
            return flatten_blocks(self.lines, num_indents=self.indent)

    def get_import_lines(self):
        """Take the stored imports and converts them to lines"""
        if self.imports:
            return [
                "from %s import %s" % (value, key)
                for key, value in self.imports.items()
            ]
        else:
            return []

    import_lines = property(get_import_lines)


class ModelCode(Code):
    """Produces a python script that can recreate data for a given model class."""

    def __init__(self, model, context=None, stdout=None, stderr=None, options=None):
        super().__init__(indent=0, stdout=stdout, stderr=stderr)
        self.model = model
        if context is None:
            context = {}
        self.context = context
        self.options = options
        self.instances = []

    def get_imports(self):
        """
        Return a dictionary of import statements, with the variable being
        defined as the key.
        """
        return {self.model.__name__: smart_str(self.model.__module__)}

    imports = property(get_imports)

    def get_lines(self):
        """
        Return a list of lists or strings, representing the code body.
        Each list is a block, each string is a statement.
        """
        code = []

        for counter, item in enumerate(self.model._default_manager.all()):
            instance = InstanceCode(
                instance=item,
                id=counter + 1,
                context=self.context,
                stdout=self.stdout,
                stderr=self.stderr,
                options=self.options,
            )
            self.instances.append(instance)
            if instance.waiting_list:
                code += instance.lines

        # After each instance has been processed, try again.
        # This allows self referencing fields to work.
        for instance in self.instances:
            if instance.waiting_list:
                code += instance.lines

        return code

    lines = property(get_lines)


class InstanceCode(Code):
    """Produces a python script that can recreate data for a given model instance."""

    def __init__(
        self, instance, id, context=None, stdout=None, stderr=None, options=None
    ):
        """We need the instance in question and an id"""

        super().__init__(indent=0, stdout=stdout, stderr=stderr)
        self.imports = {}

        self.options = options
        self.instance = instance
        self.model = self.instance.__class__
        if context is None:
            context = {}
        self.context = context
        self.variable_name = "%s_%s" % (self.instance._meta.db_table, id)
        self.skip_me = None
        self.instantiated = False

        self.waiting_list = list(self.model._meta.fields)

        self.many_to_many_waiting_list = {}
        for field in self.model._meta.many_to_many:
            try:
                if not field.remote_field.through._meta.auto_created:
                    continue
            except AttributeError:
                pass
            self.many_to_many_waiting_list[field] = list(
                getattr(self.instance, field.name).all()
            )

    def get_lines(self, force=False):
        """
        Return a list of lists or strings, representing the code body.
        Each list is a block, each string is a statement.

        force (True or False): if an attribute object cannot be included,
        it is usually skipped to be processed later. With 'force' set, there
        will be no waiting: a get_or_create() call is written instead.
        """
        code_lines = []

        # Don't return anything if this is an instance that should be skipped
        if self.skip():
            return []

        # Initialise our new object
        # e.g. model_name_35 = Model()
        code_lines += self.instantiate()

        # Add each field
        # e.g. model_name_35.field_one = 1034.91
        #      model_name_35.field_two = "text"
        code_lines += self.get_waiting_list()

        if force:
            # TODO: Check that M2M are not affected
            code_lines += self.get_waiting_list(force=force)

        # Print the save command for our new object
        # e.g. model_name_35.save()
        if code_lines:
            code_lines.append(
                "%s = importer.save_or_locate(%s)\n"
                % (self.variable_name, self.variable_name)
            )

        code_lines += self.get_many_to_many_lines(force=force)

        return code_lines

    lines = property(get_lines)

    def skip(self):
        """
        Determine whether or not this object should be skipped.
        If this model instance is a parent of a single subclassed
        instance, skip it. The subclassed instance will create this
        parent instance for us.

        TODO: Allow the user to force its creation?
        """
        if self.skip_me is not None:
            return self.skip_me

        cls = self.instance.__class__
        using = router.db_for_write(cls, instance=self.instance)
        collector = Collector(using=using)
        collector.collect([self.instance], collect_related=False)
        sub_objects = sum([list(i) for i in collector.data.values()], [])
        sub_objects_parents = [so._meta.parents for so in sub_objects]
        if [self.model in p for p in sub_objects_parents].count(True) == 1:
            # since this instance isn't explicitly created, it's variable name
            # can't be referenced in the script, so record None in context dict
            pk_name = self.instance._meta.pk.name
            key = "%s_%s" % (self.model.__name__, getattr(self.instance, pk_name))
            self.context[key] = None
            self.skip_me = True
        else:
            self.skip_me = False

        return self.skip_me

    def instantiate(self):
        """Write lines for instantiation"""
        # e.g. model_name_35 = Model()
        code_lines = []

        if not self.instantiated:
            code_lines.append("%s = %s()" % (self.variable_name, self.model.__name__))
            self.instantiated = True

            # Store our variable name for future foreign key references
            pk_name = self.instance._meta.pk.name
            key = "%s_%s" % (self.model.__name__, getattr(self.instance, pk_name))
            self.context[key] = self.variable_name

        return code_lines

    def get_waiting_list(self, force=False):
        """Add lines for any waiting fields that can be completed now."""

        code_lines = []
        skip_autofield = self.options["skip_autofield"]

        # Process normal fields
        for field in list(self.waiting_list):
            try:
                # Find the value, add the line, remove from waiting list and move on
                value = get_attribute_value(
                    self.instance,
                    field,
                    self.context,
                    force=force,
                    skip_autofield=skip_autofield,
                )
                code_lines.append(
                    "%s.%s = %s" % (self.variable_name, field.name, value)
                )
                self.waiting_list.remove(field)
            except SkipValue:
                # Remove from the waiting list and move on
                self.waiting_list.remove(field)
                continue
            except DoLater:
                # Move on, maybe next time
                continue

        return code_lines

    def get_many_to_many_lines(self, force=False):
        """Generate lines that define many to many relations for this instance."""

        lines = []

        for field, rel_items in self.many_to_many_waiting_list.items():
            for rel_item in list(rel_items):
                try:
                    pk_name = rel_item._meta.pk.name
                    key = "%s_%s" % (
                        rel_item.__class__.__name__,
                        getattr(rel_item, pk_name),
                    )
                    value = "%s" % self.context[key]
                    lines.append(
                        "%s.%s.add(%s)" % (self.variable_name, field.name, value)
                    )
                    self.many_to_many_waiting_list[field].remove(rel_item)
                except KeyError:
                    if force:
                        item_locator = orm_item_locator(rel_item)
                        self.context["__extra_imports"][rel_item._meta.object_name] = (
                            rel_item.__module__
                        )
                        lines.append(
                            "%s.%s.add( %s )"
                            % (self.variable_name, field.name, item_locator)
                        )
                        self.many_to_many_waiting_list[field].remove(rel_item)

        if lines:
            lines.append("")

        return lines


class Script(Code):
    """Produces a complete python script that can recreate data for the given apps."""

    def __init__(self, models, context=None, stdout=None, stderr=None, options=None):
        super().__init__(stdout=stdout, stderr=stderr)
        self.imports = {}

        self.models = models
        if context is None:
            context = {}
        self.context = context

        self.context["__avaliable_models"] = set(models)
        self.context["__extra_imports"] = {}

        self.options = options

    def _queue_models(self, models, context):
        """
        Work an an appropriate ordering for the models.
        This isn't essential, but makes the script look nicer because
        more instances can be defined on their first try.
        """
        model_queue = []
        number_remaining_models = len(models)
        # Max number of cycles allowed before we call it an infinite loop.
        MAX_CYCLES = number_remaining_models
        allowed_cycles = MAX_CYCLES

        while number_remaining_models > 0:
            previous_number_remaining_models = number_remaining_models

            model = models.pop(0)

            # If the model is ready to be processed, add it to the list
            if check_dependencies(model, model_queue, context["__avaliable_models"]):
                model_class = ModelCode(
                    model=model,
                    context=context,
                    stdout=self.stdout,
                    stderr=self.stderr,
                    options=self.options,
                )
                model_queue.append(model_class)

            # Otherwise put the model back at the end of the list
            else:
                models.append(model)

            # Check for infinite loops.
            # This means there is a cyclic foreign key structure
            # That cannot be resolved by re-ordering
            number_remaining_models = len(models)
            if number_remaining_models == previous_number_remaining_models:
                allowed_cycles -= 1
                if allowed_cycles <= 0:
                    # Add remaining models, but do not remove them from the model list
                    missing_models = [
                        ModelCode(
                            model=m,
                            context=context,
                            stdout=self.stdout,
                            stderr=self.stderr,
                            options=self.options,
                        )
                        for m in models
                    ]
                    model_queue += missing_models
                    # Replace the models with the model class objects
                    # (sure, this is a little bit of hackery)
                    models[:] = missing_models
                    break
            else:
                allowed_cycles = MAX_CYCLES

        return model_queue

    def get_lines(self):
        """
        Return a list of lists or strings, representing the code body.
        Each list is a block, each string is a statement.
        """
        code = [self.FILE_HEADER.strip()]

        # Queue and process the required models
        for model_class in self._queue_models(self.models, context=self.context):
            msg = "Processing model: %s.%s\n" % (
                model_class.model.__module__,
                model_class.model.__name__,
            )
            self.stderr.write(msg)
            code.append("    # " + msg)
            code.append(model_class.import_lines)
            code.append("")
            code.append(model_class.lines)

        # Process left over foreign keys from cyclic models
        for model in self.models:
            msg = "Re-processing model: %s.%s\n" % (
                model.model.__module__,
                model.model.__name__,
            )
            self.stderr.write(msg)
            code.append("    # " + msg)
            for instance in model.instances:
                if instance.waiting_list or instance.many_to_many_waiting_list:
                    code.append(instance.get_lines(force=True))

        code.insert(1, "    # Initial Imports")
        code.insert(2, "")
        for key, value in self.context["__extra_imports"].items():
            code.insert(2, "    from %s import %s" % (value, key))

        return code

    lines = property(get_lines)

    # A user-friendly file header
    FILE_HEADER = """

#!/usr/bin/env python


# This file has been automatically generated.
# Instead of changing it, create a file called import_helper.py
# and put there a class called ImportHelper(object) in it.
#
# This class will be specially cast so that instead of extending object,
# it will actually extend the class BasicImportHelper()
#
# That means you just have to overload the methods you want to
# change, leaving the other ones intact.
#
# Something that you might want to do is use transactions, for example.
#
# Also, don't forget to add the necessary Django imports.
#
# This file was generated with the following command:
# %s
#
# to restore it, run
# manage.py runscript module_name.this_script_name
#
# example: if manage.py is at ./manage.py
# and the script is at ./some_folder/some_script.py
# you must make sure ./some_folder/__init__.py exists
# and run  ./manage.py runscript some_folder.some_script
import os, sys
from django.db import transaction

class BasicImportHelper:

    def pre_import(self):
        pass

    @transaction.atomic
    def run_import(self, import_data):
        import_data()

    def post_import(self):
        pass

    def locate_similar(self, current_object, search_data):
        # You will probably want to call this method from save_or_locate()
        # Example:
        #   new_obj = self.locate_similar(the_obj, {"national_id": the_obj.national_id } )

        the_obj = current_object.__class__.objects.get(**search_data)
        return the_obj

    def locate_object(self, original_class, original_pk_name, the_class, pk_name, pk_value, obj_content):
        # You may change this function to do specific lookup for specific objects
        #
        # original_class class of the django orm's object that needs to be located
        # original_pk_name the primary key of original_class
        # the_class      parent class of original_class which contains obj_content
        # pk_name        the primary key of original_class
        # pk_value       value of the primary_key
        # obj_content    content of the object which was not exported.
        #
        # You should use obj_content to locate the object on the target db
        #
        # An example where original_class and the_class are different is
        # when original_class is Farmer and the_class is Person. The table
        # may refer to a Farmer but you will actually need to locate Person
        # in order to instantiate that Farmer
        #
        # Example:
        #   if the_class == SurveyResultFormat or the_class == SurveyType or the_class == SurveyState:
        #       pk_name="name"
        #       pk_value=obj_content[pk_name]
        #   if the_class == StaffGroup:
        #       pk_value=8

        search_data = { pk_name: pk_value }
        the_obj = the_class.objects.get(**search_data)
        #print(the_obj)
        return the_obj


    def save_or_locate(self, the_obj):
        # Change this if you want to locate the object in the database
        try:
            the_obj.save()
        except:
            print("---------------")
            print("Error saving the following object:")
            print(the_obj.__class__)
            print(" ")
            print(the_obj.__dict__)
            print(" ")
            print(the_obj)
            print(" ")
            print("---------------")

            raise
        return the_obj


importer = None
try:
    import import_helper
    # We need this so ImportHelper can extend BasicImportHelper, although import_helper.py
    # has no knowlodge of this class
    importer = type("DynamicImportHelper", (import_helper.ImportHelper, BasicImportHelper ) , {} )()
except ImportError as e:
    # From Python 3.3 we can check e.name - string match is for backward compatibility.
    if 'import_helper' in str(e):
        importer = BasicImportHelper()
    else:
        raise

import datetime
from decimal import Decimal
from django.contrib.contenttypes.models import ContentType

try:
    import dateutil.parser
    from dateutil.tz import tzoffset
except ImportError:
    print("Please install python-dateutil")
    sys.exit(os.EX_USAGE)

def run():
    importer.pre_import()
    importer.run_import(import_data)
    importer.post_import()

def import_data():

""" % " ".join(sys.argv)  # noqa: E501


# HELPER FUNCTIONS
# -------------------------------------------------------------------------------


def flatten_blocks(lines, num_indents=-1):
    """
    Take a list (block) or string (statement) and flattens it into a string
    with indentation.
    """
    # The standard indent is four spaces
    INDENTATION = " " * 4

    if not lines:
        return ""

    # If this is a string, add the indentation and finish here
    if isinstance(lines, str):
        return INDENTATION * num_indents + lines

    # If this is not a string, join the lines and recurse
    return "\n".join([flatten_blocks(line, num_indents + 1) for line in lines])


def get_attribute_value(item, field, context, force=False, skip_autofield=True):
    """Get a string version of the given attribute's value, like repr() might."""
    # Find the value of the field, catching any database issues
    try:
        value = getattr(item, field.name)
    except ObjectDoesNotExist:
        raise SkipValue(
            "Could not find object for %s.%s, ignoring.\n"
            % (item.__class__.__name__, field.name)
        )

    # AutoField: We don't include the auto fields, they'll be automatically recreated
    if skip_autofield and isinstance(field, AutoField):
        raise SkipValue()

    # Some databases (eg MySQL) might store boolean values as 0/1,
    # this needs to be cast as a bool
    elif isinstance(field, BooleanField) and value is not None:
        return repr(bool(value))

    # Post file-storage-refactor, repr() on File/ImageFields no longer returns the path
    elif isinstance(field, FileField):
        return repr(force_str(value))

    # ForeignKey fields, link directly using our stored python variable name
    elif isinstance(field, ForeignKey) and value is not None:
        # Special case for contenttype foreign keys: no need to output any
        # content types in this script, as they can be generated again
        # automatically.
        # NB: Not sure if "is" will always work
        if field.remote_field.model is ContentType:
            return 'ContentType.objects.get(app_label="%s", model="%s")' % (
                value.app_label,
                value.model,
            )

        # Generate an identifier (key) for this foreign object
        pk_name = value._meta.pk.name
        key = "%s_%s" % (value.__class__.__name__, getattr(value, pk_name))

        if key in context:
            variable_name = context[key]
            # If the context value is set to None, this should be skipped.
            # This identifies models that have been skipped (inheritance)
            if variable_name is None:
                raise SkipValue()
            # Return the variable name listed in the context
            return "%s" % variable_name
        elif value.__class__ not in context["__avaliable_models"] or force:
            context["__extra_imports"][value._meta.object_name] = value.__module__
            item_locator = orm_item_locator(value)
            return item_locator
        else:
            raise DoLater("(FK) %s.%s\n" % (item.__class__.__name__, field.name))

    elif isinstance(field, (DateField, DateTimeField)) and value is not None:
        return 'dateutil.parser.parse("%s")' % value.isoformat()

    # A normal field (e.g. a python built-in)
    else:
        return repr(value)


def make_clean_dict(the_dict):
    if "_state" in the_dict:
        clean_dict = the_dict.copy()
        del clean_dict["_state"]
        return clean_dict
    return the_dict


def check_dependencies(model, model_queue, avaliable_models):
    """Check that all the depenedencies for this model are already in the queue."""
    # A list of allowed links: existing fields, itself and the special case ContentType
    allowed_links = [m.model.__name__ for m in model_queue] + [
        model.__name__,
        "ContentType",
    ]

    # For each ForeignKey or ManyToMany field, check that a link is possible

    for field in model._meta.fields:
        if not field.remote_field:
            continue
        if field.remote_field.model.__name__ not in allowed_links:
            if field.remote_field.model not in avaliable_models:
                continue
            return False

    for field in model._meta.many_to_many:
        if not field.remote_field:
            continue
        if field.remote_field.model.__name__ not in allowed_links:
            return False

    return True


# EXCEPTIONS
# -------------------------------------------------------------------------------


class SkipValue(Exception):
    """Value could not be parsed or should simply be skipped."""


class DoLater(Exception):
    """Value could not be parsed or should simply be skipped."""


class StrToCodeChanger:
    def __init__(self, string):
        self.repr = string

    def __repr__(self):
        return self.repr



# -*- coding: utf-8 -*-
from io import StringIO

from django.core.management import call_command


def test_show_template_tags():
    out = StringIO()
    call_command("show_template_tags", "--no-color", stdout=out)

    output = out.getvalue()

    assert "App: django.contrib.admin\n" in output
    assert "App: django_extensions\n" in output
    assert "load: highlighting\n" in output
    assert "Tag: highlight\n" in output


def test_show_template_tags_testapp():
    out = StringIO()
    call_command("show_template_tags", "--no-color", stdout=out)

    output = out.getvalue()

    assert "App: tests.testapp\n" in output
    assert "load: dummy_tags\n" in output
    assert "Tag: dummy_tag\n" in output


def test_show_template_tags_testapp_with_appconfig():
    out = StringIO()
    call_command("show_template_tags", "--no-color", stdout=out)

    output = out.getvalue()

    assert "App: tests.testapp_with_appconfig\n" in output
    assert "load: dummy_tags_appconfig\n" in output
    assert "Tag: dummy_tag_appconfig\n" in output



# -*- coding: utf-8 -*-
from io import StringIO

import pytest
from django.contrib.auth.models import User
from django.core.management import CommandError, call_command

from django_extensions.management.commands.set_fake_passwords import (
    DEFAULT_FAKE_PASSWORD,
)

from unittest.mock import Mock, patch


@pytest.fixture(scope="module")
def django_db_setup(django_db_setup, django_db_blocker):
    """Load to database a set of users, create for export
    emails command testing"""
    with django_db_blocker.unblock():
        call_command("loaddata", "group.json")
        call_command("loaddata", "user.json")


@pytest.mark.django_db()
def test_without_args(capsys, settings):
    settings.DEBUG = True

    old_passwords = User.objects.values_list("password", flat=True).order_by("id")
    assert len(set(old_passwords)) == 3

    call_command("set_fake_passwords")
    out, err = capsys.readouterr()
    assert "Reset 3 passwords" in out

    new_passwords = User.objects.values_list("password", flat=True).order_by("id")
    assert len(set(new_passwords)) == 1
    assert old_passwords != new_passwords

    for user in User.objects.all():
        assert user.check_password(DEFAULT_FAKE_PASSWORD)


@pytest.mark.django_db()
def test_with_password(capsys, settings):
    settings.DEBUG = True

    call_command("set_fake_passwords", "--password=helloworld")
    out, err = capsys.readouterr()
    assert "Reset 3 passwords" in out

    for user in User.objects.all():
        assert user.check_password("helloworld")


@pytest.mark.django_db()
def test_with_prompt(settings):
    settings.DEBUG = True

    m_getpass = Mock()
    m_getpass.getpass.return_value = "test"

    with patch.dict("sys.modules", getpass=m_getpass):
        call_command("set_fake_passwords", "--prompt")

    assert all([user.check_password("test") for user in User.objects.all()])


@pytest.mark.django_db()
def test_with_prompt_with_empty_password(settings):
    settings.DEBUG = True

    m_getpass = Mock()
    m_getpass.getpass.return_value = None

    with pytest.raises(CommandError, match="You must enter a valid password"):
        with patch.dict("sys.modules", getpass=m_getpass):
            call_command("set_fake_passwords", "--prompt")


def test_without_debug(settings):
    settings.DEBUG = False

    out = StringIO()
    with pytest.raises(CommandError, match="Only available in debug mode"):
        call_command("set_fake_passwords", verbosity=3, stdout=out, stderr=out)



# -*- coding: utf-8 -*-
from unittest import mock
import os
from io import StringIO

from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase


DefaultCacheMock = mock.Mock()
OtherCacheMock = mock.Mock()


class ClearCacheTests(TestCase):
    def setUp(self):
        self.project_root = os.path.join("tests", "testapp")
        self._settings = os.environ.get("DJANGO_SETTINGS_MODULE")
        os.environ["DJANGO_SETTINGS_MODULE"] = "django_extensions.settings"
        DefaultCacheMock.reset_mock()
        OtherCacheMock.reset_mock()
        self.out = StringIO()

    def tearDown(self):
        if self._settings:
            os.environ["DJANGO_SETTINGS_MODULE"] = self._settings

    def test_called_with_no_arguments(self):
        with self.settings(BASE_DIR=self.project_root):
            call_command("clear_cache")
        DefaultCacheMock.return_value.clear.assert_called()
        OtherCacheMock.return_value.clear.assert_not_called()

    def test_called_with_explicit_other(self):
        with self.settings(BASE_DIR=self.project_root):
            call_command("clear_cache", "--cache", "other")
        DefaultCacheMock.return_value.clear.assert_not_called()
        OtherCacheMock.return_value.clear.assert_called()

    def test_called_with_all_argument(self):
        with self.settings(BASE_DIR=self.project_root):
            call_command("clear_cache", "--all")
        DefaultCacheMock.return_value.clear.assert_called()
        OtherCacheMock.return_value.clear.assert_called()

    def test_called_with_explicit_all(self):
        with self.settings(BASE_DIR=self.project_root):
            call_command(
                "clear_cache", "--cache", "default", "--cache", "other", stdout=self.out
            )
        DefaultCacheMock.return_value.clear.assert_called()
        OtherCacheMock.return_value.clear.assert_called()

        self.assertIn('Cache "default" has been cleared!', self.out.getvalue())
        self.assertIn('Cache "other" has been cleared!', self.out.getvalue())

    def test_called_with_invalid_arguments(self):
        with self.settings(BASE_DIR=self.project_root):
            with self.assertRaisesRegex(
                CommandError, "Using both --all and --cache is not supported"
            ):
                call_command("clear_cache", "--all", "--cache", "foo")

    def test_should_print_that_cache_is_invalid_on_InvalidCacheBackendError(self):
        call_command("clear_cache", "--cache", "invalid", stderr=self.out)

        self.assertIn('Cache "invalid" is invalid!', self.out.getvalue())



# -*- coding: utf-8 -*-
from django.test import TestCase
from django.template import Context, Template, TemplateSyntaxError


class IndentByTagExceptions(TestCase):
    """Test for indentby exceptions."""

    def test_should_raise_TemplateSyntaxError_if_args_lenght_not_in_2_4(self):
        content = """{% load indent_text %}
{% indentby %}
Hello World
{% endindentby %}"""
        with self.assertRaisesRegex(
            TemplateSyntaxError, "indentby tag requires 1 or 3 arguments"
        ):
            Template(content).render(Context())


class IndentByTagTests(TestCase):
    """Tests for indentby tag."""

    def test_should_add_4_spaces_indent_before_given_text(self):
        content = """{% load indent_text %}
{% indentby 4 %}
Hello World
{% endindentby %}"""
        expected_result = "\n    \n    Hello World\n"

        result = Template(content).render(Context())

        self.assertEqual(result, expected_result)

    def test_should_add_2_spaces_indent_before_given_text_if_statement_True(self):
        content = """{% load indent_text %}
{% indentby 2 if test_statement %}
Hello World
{% endindentby %}"""
        expected_result = "\n  \n  Hello World\n"

        result = Template(content).render(Context({"test_statement": True}))

        self.assertEqual(result, expected_result)

    def test_should_not_add_any_spaces_indent_before_given_text_if_statement_variable_does_not_exist(
        self,
    ):
        content = """{% load indent_text %}
{% indentby 2 if test_statement %}
Hello World
{% endindentby %}"""
        expected_result = "\n\nHello World\n"

        result = Template(content).render(Context())

        self.assertEqual(result, expected_result)



# -*- coding: utf-8 -*-
import sys
import csv

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.core.management.base import BaseCommand, CommandError

from django_extensions.management.utils import signalcommand


FORMATS = [
    "address",
    "emails",
    "google",
    "outlook",
    "linkedin",
    "vcard",
]


def full_name(**kwargs):
    """Return full name or username."""
    first_name = kwargs.get("first_name")
    last_name = kwargs.get("last_name")

    name = " ".join(n for n in [first_name, last_name] if n)
    if name:
        return name

    name = kwargs.get("name")
    if name:
        return name

    username = kwargs.get("username")
    if username:
        return username

    return ""


class Command(BaseCommand):
    help = "Export user email address list in one of a number of formats."
    args = "[output file]"
    label = "filename to save to"

    can_import_settings = True
    encoding = "utf-8"  # RED_FLAG: add as an option -DougN

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.UserModel = get_user_model()

    def add_arguments(self, parser):
        super().add_arguments(parser)
        (
            parser.add_argument(
                "--group",
                "-g",
                action="store",
                dest="group",
                default=None,
                help="Limit to users which are part of the supplied group name",
            ),
        )
        parser.add_argument(
            "--format",
            "-f",
            action="store",
            dest="format",
            default=FORMATS[0],
            help="output format. May be one of %s." % ", ".join(FORMATS),
        )

    def full_name(self, **kwargs):
        return getattr(settings, "EXPORT_EMAILS_FULL_NAME_FUNC", full_name)(**kwargs)

    @signalcommand
    def handle(self, *args, **options):
        if len(args) > 1:
            raise CommandError("extra arguments supplied")
        group = options["group"]
        if group and not Group.objects.filter(name=group).count() == 1:
            names = "', '".join(g["name"] for g in Group.objects.values("name"))
            if names:
                names = "'" + names + "'."
            raise CommandError(
                "Unknown group '" + group + "'. Valid group names are: " + names
            )

        UserModel = get_user_model()
        order_by = getattr(
            settings,
            "EXPORT_EMAILS_ORDER_BY",
            ["last_name", "first_name", "username", "email"],
        )
        fields = getattr(
            settings,
            "EXPORT_EMAILS_FIELDS",
            ["last_name", "first_name", "username", "email"],
        )

        qs = UserModel.objects.all().order_by(*order_by)
        if group:
            qs = qs.filter(groups__name=group).distinct()
        qs = qs.values(*fields)
        getattr(self, options["format"])(qs)

    def address(self, qs):
        """
        Single entry per line in the format of:
            "full name" <my@address.com>;
        """
        self.stdout.write(
            "\n".join(
                '"%s" <%s>;' % (self.full_name(**ent), ent.get("email", ""))
                for ent in qs
            )
        )
        self.stdout.write("\n")

    def emails(self, qs):
        """
        Single entry with email only in the format of:
            my@address.com,
        """
        self.stdout.write(",\n".join(ent["email"] for ent in qs if ent.get("email")))
        self.stdout.write("\n")

    def google(self, qs):
        """CSV format suitable for importing into google GMail"""
        csvf = csv.writer(sys.stdout)
        csvf.writerow(["Name", "Email"])
        for ent in qs:
            csvf.writerow([self.full_name(**ent), ent.get("email", "")])

    def linkedin(self, qs):
        """
        CSV format suitable for importing into linkedin Groups.
        perfect for pre-approving members of a linkedin group.
        """
        csvf = csv.writer(sys.stdout)
        csvf.writerow(["First Name", "Last Name", "Email"])
        for ent in qs:
            csvf.writerow(
                [
                    ent.get("first_name", ""),
                    ent.get("last_name", ""),
                    ent.get("email", ""),
                ]
            )

    def outlook(self, qs):
        """CSV format suitable for importing into outlook"""
        csvf = csv.writer(sys.stdout)
        columns = [
            "Name",
            "E-mail Address",
            "Notes",
            "E-mail 2 Address",
            "E-mail 3 Address",
            "Mobile Phone",
            "Pager",
            "Company",
            "Job Title",
            "Home Phone",
            "Home Phone 2",
            "Home Fax",
            "Home Address",
            "Business Phone",
            "Business Phone 2",
            "Business Fax",
            "Business Address",
            "Other Phone",
            "Other Fax",
            "Other Address",
        ]
        csvf.writerow(columns)
        empty = [""] * (len(columns) - 2)
        for ent in qs:
            csvf.writerow([self.full_name(**ent), ent.get("email", "")] + empty)

    def vcard(self, qs):
        """VCARD format."""
        try:
            import vobject
        except ImportError:
            print(
                self.style.ERROR(
                    "Please install vobject to use the vcard export format."
                )
            )
            sys.exit(1)

        out = sys.stdout
        for ent in qs:
            card = vobject.vCard()
            card.add("fn").value = self.full_name(**ent)
            if ent.get("last_name") and ent.get("first_name"):
                card.add("n").value = vobject.vcard.Name(
                    ent["last_name"], ent["first_name"]
                )
            else:
                # fallback to fullname, if both first and lastname are not declared
                card.add("n").value = vobject.vcard.Name(self.full_name(**ent))
            if ent.get("email"):
                emailpart = card.add("email")
                emailpart.value = ent["email"]
                emailpart.type_param = "INTERNET"

            out.write(card.serialize())



# -*- coding: utf-8 -*-
from django.core.management import call_command

from django_extensions.management.commands.export_emails import full_name

import pytest

from tests.testapp.settings import DATABASES


@pytest.fixture(autouse=True)
def custom_djsettings(settings):
    """Custom django settings to avoid warnings in stdout"""
    settings.TEMPLATE_DEBUG = False
    settings.DEBUG = False
    return settings


@pytest.fixture(scope="module")
def django_db_setup():
    """Select default database for testing"""
    settings.DATABASES = DATABASES  # noqa


@pytest.fixture(scope="module")  # noqa
def django_db_setup(django_db_setup, django_db_blocker):  # noqa
    """Load to database a set of users, create for export
    emails command testing"""
    with django_db_blocker.unblock():
        call_command("loaddata", "group.json")
        call_command("loaddata", "user.json")


@pytest.mark.django_db()
def test_do_export_emails_stdout_start(capsys):
    """Testing export_emails command without args.stdout starts."""
    call_command("export_emails")

    out, err = capsys.readouterr()
    assert out.startswith('"Mijaíl Bulgakóv')


@pytest.mark.django_db()
def test_do_export_emails_stdout_end(capsys):
    """Testing export_emails command without args.stdout end."""
    call_command("export_emails")

    out, err = capsys.readouterr()
    assert '"Frédéric Mistral" <frederic_mistral@gmail.com>;\n\n' in out


@pytest.mark.django_db()
def test_do_export_emails_format_email(capsys):
    """Testing python manage.py export_emails -f emails"""
    call_command("export_emails", "--format=emails")

    out, err = capsys.readouterr()
    assert "frederic_mistral@gmail.com" in out


@pytest.mark.django_db()
def test_do_export_emails_address(capsys):
    """Testing python manage export_emails -f address"""
    call_command("export_emails", "--format=address")

    out, err = capsys.readouterr()
    assert '"Frédéric Mistral"' in out


@pytest.mark.django_db()
def test_do_export_emails_format_google(capsys):
    """Testing python manage export_emails -f google"""
    call_command("export_emails", "--format=google")

    out, err = capsys.readouterr()
    assert out.startswith("Name,Email")


@pytest.mark.django_db()
def test_do_export_emails_format_linkedin(capsys):
    """Testing python manage.py export_emails -f linkedin"""
    call_command("export_emails", "--format=linkedin")

    out, err = capsys.readouterr()

    assert out.startswith("First Name,")
    assert "Gabriel Garcia,Marquéz" in out


@pytest.mark.django_db()
def test_do_export_emails_format_outlook(capsys):
    """Testing python manage.py export_emails -f outlook"""
    call_command("export_emails", "--format=outlook")

    out, err = capsys.readouterr()
    assert out.startswith("Name,E-mail Address")
    assert "frederic_mistral@gmail.com" in out


@pytest.mark.django_db()
def test_do_export_emails_format_vcard_start(capsys):
    """Testing python manage.py export_emails -f vcard"""
    call_command("export_emails", "--format=vcard")

    out, err = capsys.readouterr()

    assert "N:Bulgakóv;Mijaíl;;;" in out
    assert out.startswith("BEGIN:VCARD")


@pytest.mark.django_db()
def test_full_name():
    """Test, getting full name / username"""

    fake_user = {"first_name": "Allan", "last_name": "Poe", "username": "allan_poe"}
    name = full_name(**fake_user)
    assert name == "{fn} {ln}".format(
        fn=fake_user["first_name"], ln=fake_user["last_name"]
    )

    fake_user2 = {"first_name": "", "last_name": "", "username": "niccolas"}
    name2 = full_name(**fake_user2)
    assert name2 == fake_user2["username"]



# -*- coding: utf-8 -*-
from django.core import mail
from django.core.management import call_command
from django.test import TestCase
from django.test.utils import override_settings
from io import StringIO

from unittest.mock import patch


class EmailNotificationCommandTests(TestCase):
    """Tests for EmailNotificationCommand class."""

    @override_settings(ADMINS=[])
    @patch("sys.stdout", new_callable=StringIO)
    def test_should_print_that_no_email_recipients_available(self, m_stdout):
        with self.assertRaises(Exception):
            call_command(
                "test_email_notification_command", "--email-exception", verbosity=2
            )

        self.assertIn("No email recipients available", m_stdout.getvalue())
        self.assertListEqual(mail.outbox, [])

    @override_settings(
        ADMINS=["foo@bar.com", "bar@foo.com"], DEFAULT_FROM_EMAIL="webmaster@foo.bar"
    )
    def test_should_send_email_with_command_name_and_full_traceback_if_command_fail(
        self,
    ):
        expected_lines = """Reporting execution of command: 'test_email_notification_command'
Traceback:
    raise Exception()"""

        with self.assertRaises(Exception):
            call_command(
                "test_email_notification_command", "--email-exception", verbosity=2
            )

        self.assertIsNot(mail.outbox, [])
        self.assertEqual(
            mail.outbox[0].subject, "Django extensions email notification."
        )
        self.assertEqual(mail.outbox[0].from_email, "webmaster@foo.bar")
        self.assertListEqual(mail.outbox[0].to, ["foo@bar.com", "bar@foo.com"])
        for expected_line in expected_lines.splitlines():
            self.assertIn(expected_line, mail.outbox[0].body)

    @patch("sys.stdout", new_callable=StringIO)
    def test_should_not_notify_if_notification_level_is_not_set(self, m_stdout):
        call_command("runscript", "sample_script", "--email-notifications", verbosity=2)

        self.assertIn("Exiting, not in 'notify always' mode", m_stdout.getvalue())
        self.assertListEqual(mail.outbox, [])

    @override_settings(
        ADMINS=["foo@bar.com"],
        DEFAULT_FROM_EMAIL="webmaster@foo.bar",
        EMAIL_NOTIFICATIONS={
            "tests.testapp.scripts.sample_script": {
                "subject": "my_script subject",
                "body": "my_script body",
                "from_email": "from_email@example.com",
                "recipients": ("recipient0@example.com",),
                "no_admins": False,
                "no_traceback": False,
                "notification_level": 1,
                "fail_silently": False,
            }
        },
    )
    def test_should_notify_if_notification_level_is_greater_than_0(self):
        call_command("runscript", "sample_script", "--email-notifications", verbosity=2)

        self.assertIsNot(mail.outbox, [])
        self.assertEqual(mail.outbox[0].subject, "my_script subject")
        self.assertEqual(mail.outbox[0].body, "my_script body")
        self.assertEqual(mail.outbox[0].from_email, "from_email@example.com")
        self.assertListEqual(
            mail.outbox[0].to, ["recipient0@example.com", "foo@bar.com"]
        )



# -*- coding: utf-8 -*-
import os
import sys
import importlib
from typing import Optional  # NOQA
from django.apps import apps

_jobs = None


def noneimplementation(meth):
    return None


class JobError(Exception):
    pass


class BaseJob:
    help = "undefined job description."
    when = None  # type: Optional[str]

    def execute(self):
        raise NotImplementedError("Job needs to implement the execute method")


class MinutelyJob(BaseJob):
    when = "minutely"


class QuarterHourlyJob(BaseJob):
    when = "quarter_hourly"


class HourlyJob(BaseJob):
    when = "hourly"


class DailyJob(BaseJob):
    when = "daily"


class WeeklyJob(BaseJob):
    when = "weekly"


class MonthlyJob(BaseJob):
    when = "monthly"


class YearlyJob(BaseJob):
    when = "yearly"


def my_import(name):
    try:
        imp = __import__(name)
    except ImportError as err:
        raise JobError("Failed to import %s with error %s" % (name, err))

    mods = name.split(".")
    if len(mods) > 1:
        for mod in mods[1:]:
            imp = getattr(imp, mod)
    return imp


def find_jobs(jobs_dir):
    try:
        return sorted(
            [
                f[:-3]
                for f in os.listdir(jobs_dir)
                if not f.startswith("_") and f.endswith(".py")
            ]
        )
    except OSError:
        return []


def find_job_module(app_name: str, when: Optional[str] = None) -> str:
    """Find the directory path to a job module."""
    parts = app_name.split(".")
    parts.append("jobs")
    if when:
        parts.append(when)
    module_name = ".".join(parts)
    module = importlib.import_module(module_name)

    if not hasattr(module, "__path__"):
        # module here is a non-package module, eg jobs.py
        raise ImportError

    return module.__path__[0]


def import_job(app_name, name, when=None):
    jobmodule = "%s.jobs.%s%s" % (app_name, when and "%s." % when or "", name)
    job_mod = my_import(jobmodule)
    # todo: more friendly message for AttributeError if job_mod does not exist
    try:
        job = job_mod.Job
    except AttributeError:
        raise JobError(
            "Job module %s does not contain class instance named 'Job'" % jobmodule
        )
    if when and not (job.when == when or job.when is None):
        raise JobError("Job %s is not a %s job." % (jobmodule, when))
    return job


def get_jobs(when=None, only_scheduled=False):
    """
    Return a dictionary mapping of job names together with their respective
    application class.
    """
    # FIXME: HACK: make sure the project dir is on the path when executed as ./manage.py
    try:
        cpath = os.path.dirname(os.path.realpath(sys.argv[0]))
        ppath = os.path.dirname(cpath)
        if ppath not in sys.path:
            sys.path.append(ppath)
    except Exception:
        pass
    _jobs = {}

    for app_name in [app.name for app in apps.get_app_configs()]:
        scandirs = (
            None,
            "minutely",
            "quarter_hourly",
            "hourly",
            "daily",
            "weekly",
            "monthly",
            "yearly",
        )
        if when:
            scandirs = None, when
        for subdir in scandirs:
            try:
                path = find_job_module(app_name, subdir)
                for name in find_jobs(path):
                    if (app_name, name) in _jobs:
                        raise JobError("Duplicate job %s" % name)
                    job = import_job(app_name, name, subdir)
                    if only_scheduled and job.when is None:
                        # only include jobs which are scheduled
                        continue
                    if when and job.when != when:
                        # generic job not in same schedule
                        continue
                    _jobs[(app_name, name)] = job
            except ImportError:
                # No job module -- continue scanning
                pass

    return _jobs


def get_job(app_name, job_name):
    jobs = get_jobs()
    if app_name:
        return jobs[(app_name, job_name)]
    else:
        for a, j in jobs.keys():
            if j == job_name:
                return jobs[(a, j)]
        raise KeyError("Job not found: %s" % job_name)


def print_jobs(
    when=None,
    only_scheduled=False,
    show_when=True,
    show_appname=False,
    show_header=True,
):
    jobmap = get_jobs(when, only_scheduled=only_scheduled)
    print("Job List: %i jobs" % len(jobmap))
    jlist = sorted(jobmap.keys())
    if not jlist:
        return

    appname_spacer = "%%-%is" % max(len(e[0]) for e in jlist)
    name_spacer = "%%-%is" % max(len(e[1]) for e in jlist)
    when_spacer = "%%-%is" % max(len(e.when) for e in jobmap.values() if e.when)
    if show_header:
        line = " "
        if show_appname:
            line += appname_spacer % "appname" + " - "
        line += name_spacer % "jobname"
        if show_when:
            line += " - " + when_spacer % "when"
        line += " - help"
        print(line)
        print("-" * 80)

    for app_name, job_name in jlist:
        job = jobmap[(app_name, job_name)]
        line = " "
        if show_appname:
            line += appname_spacer % app_name + " - "
        line += name_spacer % job_name
        if show_when:
            line += " - " + when_spacer % (job.when and job.when or "")
        line += " - " + job.help
        print(line)



# -*- coding: utf-8 -*-
import re
from io import StringIO

from django.test import TestCase
from django.core.management import call_command


class ListSignalsTests(TestCase):
    """Tests for list_signals command."""

    def setUp(self):
        self.out = StringIO()

    def test_should_print_all_signals(self):
        expected_result = """django.contrib.sites.models.Site (site)
    pre_delete
        django.contrib.sites.models.clear_site_cache #
    pre_save
        django.contrib.sites.models.clear_site_cache #
tests.testapp.models.HasOwnerModel (has owner model)
    pre_save
        tests.testapp.models.dummy_handler #
"""

        call_command("list_signals", stdout=self.out)

        # Strip line numbers to make the test less brittle
        out = re.sub(r"(?<=#)\d+", "", self.out.getvalue(), re.M)
        self.assertIn(expected_result, out)



# -*- coding: utf-8 -*-
import fnmatch
import os
import shutil

from io import StringIO
from django.core.management import call_command
from django.test import TestCase


class CleanPycTests(TestCase):
    def setUp(self):
        self.project_root = os.path.join("tests", "testapp")
        self._settings = os.environ.get("DJANGO_SETTINGS_MODULE")
        os.environ["DJANGO_SETTINGS_MODULE"] = "django_extensions.settings"

    def tearDown(self):
        if self._settings:
            os.environ["DJANGO_SETTINGS_MODULE"] = self._settings

    def _find_pyc(self, path):
        pyc_glob = []
        for root, dirnames, filenames in os.walk(path):
            for filename in fnmatch.filter(filenames, "*.pyc"):
                pyc_glob.append(os.path.join(root, filename))
        return pyc_glob

    def test_removes_pyc_files(self):
        with self.settings(BASE_DIR=self.project_root):
            call_command("compile_pyc")
        pyc_glob = self._find_pyc(self.project_root)
        self.assertTrue(len(pyc_glob) > 0)
        with self.settings(BASE_DIR=self.project_root):
            call_command("clean_pyc")
        pyc_glob = self._find_pyc(self.project_root)
        self.assertEqual(len(pyc_glob), 0)

    def test_takes_path(self):
        out = StringIO()
        project_root = os.path.join("tests", "testapp")
        call_command("compile_pyc", path=project_root)
        pyc_glob = self._find_pyc(project_root)
        self.assertTrue(len(pyc_glob) > 0)
        call_command("clean_pyc", verbosity=2, path=project_root, stdout=out)
        output = out.getvalue().splitlines()
        self.assertEqual(sorted(pyc_glob), sorted(output))

    def test_removes_pyo_files(self):
        out = StringIO()
        project_root = os.path.join("tests", "testapp")
        call_command("compile_pyc", path=project_root)
        pyc_glob = self._find_pyc(project_root)
        self.assertTrue(len(pyc_glob) > 0)
        # Create some fake .pyo files since we can't force them to be created.
        pyo_glob = []
        for fn in pyc_glob:
            pyo = "%s.pyo" % os.path.splitext(fn)[0]
            shutil.copyfile(fn, pyo)
            pyo_glob.append(pyo)
        call_command(
            "clean_pyc", verbosity=2, path=project_root, optimize=True, stdout=out
        )
        output = out.getvalue().splitlines()
        self.assertEqual(sorted(pyc_glob + pyo_glob), sorted(output))



# -*- coding: utf-8 -*-

from django.test import TestCase

from .testapp.models import (
    ShortUUIDTestAgregateModel,
    ShortUUIDTestManyToManyModel,
    ShortUUIDTestModel_field,
    ShortUUIDTestModel_pk,
)


class ShortUUIDFieldTest(TestCase):
    def test_UUID_field_create(self):
        j = ShortUUIDTestModel_field.objects.create(
            a=6, uuid_field="vytxeTZskVKR7C7WgdSP3d"
        )
        self.assertEqual(j.uuid_field, "vytxeTZskVKR7C7WgdSP3d")

    def test_UUID_field_pk_create(self):
        j = ShortUUIDTestModel_pk.objects.create(uuid_field="vytxeTZskVKR7C7WgdSP3d")
        self.assertEqual(j.uuid_field, "vytxeTZskVKR7C7WgdSP3d")
        self.assertEqual(j.pk, "vytxeTZskVKR7C7WgdSP3d")

    def test_UUID_field_pk_agregate_create(self):
        j = ShortUUIDTestAgregateModel.objects.create(a=6)
        self.assertEqual(j.a, 6)
        self.assertIsInstance(j.pk, str)
        self.assertTrue(len(j.pk) < 23)

    def test_UUID_field_manytomany_create(self):
        j = ShortUUIDTestManyToManyModel.objects.create(
            uuid_field="vytxeTZskVKR7C7WgdSP3e"
        )
        self.assertEqual(j.uuid_field, "vytxeTZskVKR7C7WgdSP3e")
        self.assertEqual(j.pk, "vytxeTZskVKR7C7WgdSP3e")
